#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
The main device class for the PFA-2 filter from Precision Filter Inc..

The PFA-2 filter is a high-precision, low-noise analog filter commonly used in
laboratory instrumentation and signal conditioning setups. It is suitable for
applications requiring clean signal output, such as spectroscopy, low-level
voltage measurements, signal recovery and high-resolution data acquisition.

The class PFA2Filter has been tested with a PFA-2 (with option H) unit.
Verify compatibility with your specific filter variant before use,
especially for units with custom or user-modified settings.

Product information and technical details:
https://pfinc.com/product/precision-pfa-2-filter-amplifier
"""

import logging

from hvl_ccb import configdataclass
from hvl_ccb.dev.base import SingleCommDevice
from hvl_ccb.utils.validation import validate_number

from .base import (
    HPF_FREQUENCY_MAX,
    HPF_FREQUENCY_MIN,
    LPF_FREQUENCY_MAX,
    LPF_FREQUENCY_MIN,
    N_CHANNELS,
    Pfa2FilterChannelCoupling,
    Pfa2FilterChannelMode,
    Pfa2FilterCommunicationError,
    Pfa2FilterError,
    Pfa2FilterHPFState,
    Pfa2FilterLPFMode,
    Pfa2FilterOverloadMode,
    Pfa2FilterPostGain,
    Pfa2FilterPreGain,
    Pfa2FilterSerialCommunication,
    _Pfa2FilterCommands,
)

logger = logging.getLogger(__name__)


@configdataclass
class Pfa2FilterConfig:
    """Init configuration of a filter channel."""

    # Define whether the filter has the high pass filter option, default is True
    device_option_hpf: bool = True

    # Default values for both filter channels
    # Coupling mode, default is DC
    coupling_init: Pfa2FilterChannelCoupling | str = Pfa2FilterChannelCoupling.DC
    # Input mode, default is OPERATE
    mode_init: Pfa2FilterChannelMode | str = Pfa2FilterChannelMode.OPERATE
    # Pre filter gain, default is 1
    pregain_init: Pfa2FilterPreGain | float = Pfa2FilterPreGain.ONE
    # Post filter gain, default is 1
    postgain_init: Pfa2FilterPostGain | float = Pfa2FilterPostGain.ONE
    # Low pass filter frequency, default is 127750 Hz
    lpf_freq_init: int = LPF_FREQUENCY_MAX
    # Low pass filter mode, default is FLAT
    lpf_mode_init: Pfa2FilterLPFMode | str = Pfa2FilterLPFMode.FLAT
    # High pass filter state, default is ON
    hpf_state_init: bool = True
    # High pass filter frequency, default is 1 Hz
    hpf_freq_init: int = HPF_FREQUENCY_MIN
    # Overload limit, default is 10 V
    overload_limit_init: float = 10
    # Overload handling mode, default is CONTINUOUS
    overload_mode_init: Pfa2FilterOverloadMode | str = Pfa2FilterOverloadMode.CONTINUOUS


class Pfa2Filter(SingleCommDevice):
    """Device class to control the PFA-2 filter"""

    def __init__(self, com, dev_config=None) -> None:
        super().__init__(com, dev_config)

        self.channel_dict: dict[int, Channel] = {
            index: Channel(index, self.com, self.config.device_option_hpf)
            for index in range(N_CHANNELS)
        }

    def __str__(self) -> str:
        return "PFA-2 Filter"

    @staticmethod
    def config_cls() -> type[Pfa2FilterConfig]:
        return Pfa2FilterConfig

    @staticmethod
    def default_com_cls() -> type[Pfa2FilterSerialCommunication]:
        return Pfa2FilterSerialCommunication

    def start(self) -> None:
        """Starting the PFA-2 Filter"""

        logger.info(f"Starting device {self}")
        super().start()

        for channel in self.channel_dict.values():
            channel.coupling = self.config.coupling_init
            channel.input_mode = self.config.mode_init
            channel.pregain = self.config.pregain_init
            channel.postgain = self.config.postgain_init
            channel.lpf_freq = self.config.lpf_freq_init
            channel.lpf_mode = self.config.lpf_mode_init
            channel.overload_limit = self.config.overload_limit_init
            channel.overload_mode = self.config.overload_mode_init
            if self.config.device_option_hpf:
                channel.hpf_state = self.config.hpf_state_init
                channel.hpf_freq = self.config.hpf_freq_init

        logger.info("Initial settings applied")

    def stop(self) -> None:
        """
        Stop this device and close the communication protocol.
        """
        if not self.com.is_open:
            logger.warning(f"Device {self} already stopped")
            return

        logger.info(f"Stopping device {self}")
        super().stop()


class Channel:
    """
    Class to control a single channel of the PFA-2 filter.
    Will be used for each channel.
    """

    def __init__(
        self, channel: int, com: Pfa2FilterSerialCommunication, option_hpf: bool
    ) -> None:
        self._com: Pfa2FilterSerialCommunication = com
        self._channel: int = channel
        self._option_hpf: bool = option_hpf

    @property
    def coupling(self) -> Pfa2FilterChannelCoupling:
        """
        Get the coupling of the channel

        AC: AC coupling
        DC: DC coupling

        :raises Pfa2FilterCommunicationError: When the device does not respond
        :return: Channel coupling as Pfa2FilterChannelCoupling
        """
        value = self._com.query(_Pfa2FilterCommands.COUPLING.build_str(self._channel))  # type: ignore[attr-defined]
        if value is None:
            err_msg = "The device did not respond with a valid coupling value"
            logger.error(err_msg)
            raise Pfa2FilterCommunicationError(err_msg)

        value = Pfa2FilterChannelCoupling(value)
        logger.info(f"Coupling for channel {self._channel} is '{value.name}'")
        return value

    @coupling.setter
    def coupling(self, value: Pfa2FilterChannelCoupling | str) -> None:
        """
        Set the AC/DC coupling for the channel

        :param value: Value from Pfa2FilterChannelCoupling
        :raises TypeError: When value is not a string
        :raises ValueError: When value is not in Pfa2FilterChannelCoupling
        :raises Pfa2FilterCommunicationError: When the device does not respond
        """
        if not isinstance(value, str):
            raise TypeError
        if value not in Pfa2FilterChannelCoupling:  # type: ignore[attr-defined]
            raise ValueError

        response = self._com.query(
            _Pfa2FilterCommands.COUPLING.build_str(self._channel, value)  # type: ignore[attr-defined]
        )
        if response is None:
            err_msg = "The device did not respond with a valid coupling value"
            logger.error(err_msg)
            raise Pfa2FilterCommunicationError(err_msg)

        response = Pfa2FilterChannelCoupling(response)
        logger.info(
            f"Setting coupling to '{response.name}' for channel {self._channel}"
        )

    @property
    def input_mode(self) -> Pfa2FilterChannelMode:
        """
        Get the input mode

        OPERATE: Channel is operational
        CALIBRATION: Channel routed to calibration input
        SHORT: Channel is shorted to ground

        :raises Pfa2FilterCommunicationError: When input mode is not set or unknown
        :return: Input mode as Pfa2FilterChannelMode
        """
        value = self._com.query(_Pfa2FilterCommands.MODE.build_str(self._channel))  # type: ignore[attr-defined]
        if value is None:
            err_msg = f"Input mode for channel {self._channel} is not set or unknown"
            logger.error(err_msg)
            raise Pfa2FilterCommunicationError(err_msg)

        value = Pfa2FilterChannelMode(value)
        logger.debug(f"Input mode for channel {self._channel} is '{value}'")
        return value

    @input_mode.setter
    def input_mode(self, value: Pfa2FilterChannelMode | str) -> None:
        """
        Sets the input mode

        :param value: Value from Pfa2FilterChannelMode
        :raises TypeError: When value is not a string
        :raises ValueError: When value is not in Pfa2FilterChannelMode
        :raises Pfa2FilterCommunicationError: When the device does not respond
        """
        if not isinstance(value, str):
            raise TypeError
        if value not in Pfa2FilterChannelMode:  # type: ignore[attr-defined]
            raise ValueError

        response = self._com.query(
            _Pfa2FilterCommands.MODE.build_str(self._channel, value)  # type: ignore[attr-defined]
        )
        if response is None:
            err_msg = f"Input mode for channel {self._channel} is not set or unknown"
            logger.error(err_msg)
            raise Pfa2FilterCommunicationError(err_msg)

        response = Pfa2FilterChannelMode(response)
        logger.info(f"Setting input mode to '{response}' for channel {self._channel}")

    @property
    def lpf_mode(self) -> Pfa2FilterLPFMode:
        """
        Get the operation mode of the low pass filter

        OFF: Filter is disabled
        PULSE: For measurements if time domain
        FLAT: For measurements in frequency domain

        :raise Pfa2FilterCommunicationError: When LPF mode is not set or unknown
        :return: LPF filter mode as Pfa2FilterLPFMode
        """
        response = self._com.query(
            _Pfa2FilterCommands.LPF_TYPE.build_str(self._channel)  # type: ignore[attr-defined]
        )
        if response is None:
            err_msg = f"LPF mode for channel {self._channel} is not set or unknown"
            logger.error(err_msg)
            raise Pfa2FilterCommunicationError(err_msg)

        response = Pfa2FilterLPFMode(response)  # type: ignore[attr-defined]
        logger.debug(f"LPF mode for channel {self._channel} is '{response}'")
        return response

    @lpf_mode.setter
    def lpf_mode(self, value: Pfa2FilterLPFMode | str) -> None:
        """
        Set the operation mode of the low pass filter

        :param value: Value from Pfa2FilterLPFMode
        :raises TypeError: When value is not a string
        :raises ValueError: When value is not in Pfa2FilterLPFMode
        :raises Pfa2FilterCommunicationError: When LPF mode could not be set
        """
        if not isinstance(value, str):
            raise TypeError
        if value not in Pfa2FilterLPFMode:  # type: ignore[attr-defined]
            raise ValueError

        response = self._com.query(
            _Pfa2FilterCommands.LPF_TYPE.build_str(self._channel, value)  # type: ignore[attr-defined]
        )
        if response is None:
            err_msg = f"LPF mode for channel {self._channel} is not set or unknown"
            logger.error(err_msg)
            raise Pfa2FilterCommunicationError(err_msg)

        response = Pfa2FilterLPFMode(response)  # type: ignore[attr-defined]
        logger.info(f"Setting LPF mode to '{response}' for channel {self._channel}")

    @property
    def lpf_freq(self) -> int:
        """
        Get the set 3 dB cut off frequency of the low pass filter

        :raises Pfa2FilterCommunicationError: When the device does not respond
        :return: Frequency in Hz
        """
        response_str = self._com.query(
            _Pfa2FilterCommands.LPF_FREQUENCY.build_str(self._channel)  # type: ignore[attr-defined]
        )
        if response_str is None:
            err_msg = f"LPF frequency for channel {self._channel} is unknown"
            logger.error(err_msg)
            raise Pfa2FilterCommunicationError(err_msg)

        response = int(response_str)
        logger.debug(f"LPF frequency for channel {self._channel} is {response} Hz")
        return response

    @lpf_freq.setter
    def lpf_freq(self, value: int) -> None:
        """
        Set the 3 dB cut of frequency for the low pass filter

        :param value: Frequency in Hz
        :raises Pfa2FilterCommunicationError: When the device does not respond
        """
        value = self._clean_lpf_value(value)

        response_str = self._com.query(
            _Pfa2FilterCommands.LPF_FREQUENCY.build_str(self._channel, str(value))  # type: ignore[attr-defined]
        )
        if response_str is None:
            err_msg = f"LPF frequency for channel {self._channel} could not be set"
            logger.error(err_msg)
            raise Pfa2FilterCommunicationError(err_msg)

        response = int(response_str)
        logger.info(
            f"The LPF frequency for channel {self._channel} is set to {response} Hz"
        )

    @property
    def hpf_freq(self) -> int:
        """
        Get the set 3 dB cut of frequency of the high pass filter

        :raises Pfa2FilterCommunicationError: When the device does not respond
        :raises Pfa2FilterError: When the filter does not have a HPF
        :return: Frequency in Hz
        """
        if self._option_hpf:
            response_str = self._com.query(
                _Pfa2FilterCommands.HPF_FREQUENCY.build_str(self._channel)  # type: ignore[attr-defined]
            )
            if response_str is None:
                err_msg = f"HPF frequency for channel {self._channel} is unknown"
                logger.error(err_msg)
                raise Pfa2FilterCommunicationError(err_msg)

            response = int(response_str)
            logger.debug(f"HPF frequency for channel {self._channel} is {response} Hz")
            return response

        msg = "This filter does not have the high pass filter option"
        logger.error(msg)
        raise Pfa2FilterError(msg)

    @hpf_freq.setter
    def hpf_freq(self, value: int) -> None:
        """
        Set the 3 dB cut of frequency of the high pass filter

        :param value: Frequency in Hz
        :raises Pfa2FilterCommunicationError: When the device does not respond
        :raises Pfa2FilterError: When the filter does not have a HPF
        """
        if self._option_hpf:
            value = self._clean_hpf_value(value)

            response_str = self._com.query(
                _Pfa2FilterCommands.HPF_FREQUENCY.build_str(self._channel, str(value))  # type: ignore[attr-defined]
            )
            if response_str is None:
                err_msg = f"HPF frequency for channel {self._channel} could not be set"
                logger.error(err_msg)
                raise Pfa2FilterCommunicationError(err_msg)

            response = int(response_str)
            logger.info(
                f"The HPF frequency for channel {self._channel} is set to {response} Hz"
            )
            return

        msg = "This filter does not have the high pass filter option"
        logger.error(msg)
        raise Pfa2FilterError(msg)

    @property
    def hpf_state(self) -> bool:
        """
        Get the ON/OFF state of the high pass filter

        :raises Pfa2FilterCommunicationError: When the device does not respond
        :raises Pfa2FilterError: When the filter does not have a HPF
        :return: ON/OFF state of the high pass filter as bool
        """
        if self._option_hpf:
            response = (
                self._com.query(_Pfa2FilterCommands.HPF_STATE.build_str(self._channel))  # type: ignore[attr-defined]
            )
            if response is None:
                err_msg = f"HPF state for channel {self._channel} is unknown"
                logger.error(err_msg)
                raise Pfa2FilterCommunicationError(err_msg)

            response = Pfa2FilterHPFState(response)
            logger.info(f"HPF state for channel {self._channel} is {response.name}")
            return response == Pfa2FilterHPFState.ON

        msg = "This filter does not have the high pass filter option"
        logger.error(msg)
        raise Pfa2FilterError(msg)

    @hpf_state.setter
    def hpf_state(self, value: bool) -> None:
        """
        Set the ON/OFF state of the high pass filter

        :param value: Bool indicating the state
        :raises TypeError: When value is not a boolean
        :raises Pfa2FilterCommunicationError: When the device does not respond
        :raises Pfa2FilterError: When the filter does not have a HPF
        """
        if self._option_hpf:
            if not isinstance(value, bool):
                msg = "Value must be a boolean"
                raise TypeError(msg)
            msg = Pfa2FilterHPFState.ON if value else Pfa2FilterHPFState.OFF

            response = self._com.query(
                _Pfa2FilterCommands.HPF_STATE.build_str(self._channel, msg)  # type: ignore[attr-defined]
            )
            if response is None:
                err_msg = f"HPF state for channel {self._channel} could not be set"
                logger.error(err_msg)
                raise Pfa2FilterCommunicationError(err_msg)

            response = Pfa2FilterHPFState(response)
            logger.info(
                f"Setting HPF state to '{response.name}' for channel {self._channel}"
            )
            return

        err_msg = "This filter does not have the high pass filter option"
        logger.error(err_msg)
        raise Pfa2FilterError(err_msg)

    @property
    def pregain(self) -> Pfa2FilterPreGain:
        """
        Get the gain value of the pre-filter amplifier

        :raises Pfa2FilterCommunicationError: When pre filter gain is unknown
        :return: Gain value from Pfa2FilterPreGain
        """
        response = self._com.query(_Pfa2FilterCommands.PREGAIN.build_str(self._channel))  # type: ignore[attr-defined]
        if response is None:
            err_msg = f"Pre filter gain for channel {self._channel} is unknown"
            logger.error(err_msg)
            raise Pfa2FilterCommunicationError(err_msg)

        response = Pfa2FilterPreGain(float(response))
        logger.info(f"Pre filter gain for channel {self._channel} is '{response.name}'")
        return response

    @pregain.setter
    def pregain(self, value: Pfa2FilterPreGain | float) -> None:
        """
        Set the gain of the pre-filter amplifier

        :param value: Amplification value from Pfa2FilterPreGain
        :raises Pfa2FilterCommunicationError: When pre filter gain could not be set
        """
        value = Pfa2FilterPreGain(value)

        response = self._com.query(
            _Pfa2FilterCommands.PREGAIN.build_str(self._channel, str(value))  # type: ignore[attr-defined]
        )
        if response is None:
            err_msg = f"Pre filter gain for channel {self._channel} could not be set"
            logger.error(err_msg)
            raise Pfa2FilterCommunicationError(err_msg)

        response = Pfa2FilterPreGain(float(response))
        logger.info(
            f"Setting pre filter gain to {value.name} for channel {self._channel}"
        )

    @property
    def postgain(self) -> Pfa2FilterPostGain:
        """
        Get the set gain of the post filter amplifier

        :raises Pfa2FilterCommunicationError: When post filter gain is unknown
        :return: Gain value from Pfa2FilterPostGain
        """
        response = self._com.query(
            _Pfa2FilterCommands.POSTGAIN.build_str(self._channel)  # type: ignore[attr-defined]
        )
        if response is None:
            err_msg = f"Post filter gain for channel {self._channel} is unknown"
            logger.error(err_msg)
            raise Pfa2FilterCommunicationError(err_msg)

        response = Pfa2FilterPostGain(float(response))
        logger.info(
            f"Post filter gain for channel {self._channel} is '{response.name}'"
        )
        return response

    @postgain.setter
    def postgain(self, value: Pfa2FilterPostGain | float) -> None:
        """
        Set the gain of the post filter amplifier

        :param value: Amplification value from Pfa2FilterPostGain
        :raises Pfa2FilterCommunicationError: When post filter gain could not be set
        """
        value = Pfa2FilterPostGain(value)

        response = self._com.query(
            _Pfa2FilterCommands.POSTGAIN.build_str(self._channel, str(value))  # type: ignore[attr-defined]
        )
        if response is None:
            err_msg = f"Post filter gain for channel {self._channel} could not be set"
            logger.error(err_msg)
            raise Pfa2FilterCommunicationError(err_msg)

        response = Pfa2FilterPostGain(float(response))
        logger.info(
            f"Setting post filter gain to {response.name} for channel {self._channel}"
        )

    @property
    def input_overload(self) -> bool:
        """
        Check if the input is overloaded

        :raises Pfa2FilterCommunicationError: When input overload status is unknown
        :return: Bool indicating overload
        """
        response_str = self._com.query(
            _Pfa2FilterCommands.IN_OVERLOAD.build_str(self._channel)  # type: ignore[attr-defined]
        )
        if response_str is None:
            err_msg = f"Input overload for channel {self._channel} is unknown"
            logger.error(err_msg)
            raise Pfa2FilterCommunicationError(err_msg)

        response = response_str == "1"
        logger.info(
            f"Input for channel {self._channel} is"
            f"{' ' if response else ' not '}overloaded"
        )
        return response

    @property
    def output_overload(self) -> bool:
        """
        Check if the output is overloaded

        :raises Pfa2FilterCommunicationError: When output overload is unknown
        :return: Bool indicating overload
        """
        response_str = self._com.query(
            _Pfa2FilterCommands.OUT_OVERLOAD.build_str(self._channel)  # type: ignore[attr-defined]
        )
        if response_str is None:
            err_msg = f"Output overload for channel {self._channel} is unknown"
            logger.error(err_msg)
            raise Pfa2FilterCommunicationError(err_msg)

        response = response_str == "1"
        logger.info(
            f"Output for channel {self._channel} "
            f"is{' ' if response else ' not '}overloaded"
        )
        return response

    @property
    def overload_limit(self) -> float:
        """
        Get the overload limit

        :raises Pfa2FilterCommunicationError: When overload limit is unknown
        :return: Overload limit in V
        """
        response = self._com.query(
            _Pfa2FilterCommands.OVERLOAD_LIMIT.build_str(self._channel)  # type: ignore[attr-defined]
        )
        if response is None:
            err_msg = f"Overload limit for channel {self._channel} is unknown"
            logger.error(err_msg)
            raise Pfa2FilterCommunicationError(err_msg)

        value = float(response)
        logger.info(f"Overload limit for channel {self._channel} is '{value} V'")
        return value

    @overload_limit.setter
    def overload_limit(self, value: float) -> None:
        """
        Set the overload limit

        :param value: Overload limit in V in 0.1 V steps
        :raises ValueError: When value is not in [0.1 V; 10 V]
        :raises Pfa2FilterCommunicationError: When overload limit could not be set
        """
        validate_number("overload limit", value, (0.1, 10), logger=logger)

        response_str = self._com.query(
            _Pfa2FilterCommands.OVERLOAD_LIMIT.build_str(self._channel, str(value))  # type: ignore[attr-defined]
        )
        if response_str is None:
            err_msg = f"Overload limit for channel {self._channel} could not be set"
            logger.error(err_msg)
            raise Pfa2FilterCommunicationError(err_msg)

        response = float(response_str)
        logger.info(
            f"Setting overload limit to {response} V for channel {self._channel}"
        )

    @property
    def overload_mode(self) -> Pfa2FilterOverloadMode:
        """
        Get the set overload handling mode

        CONTINUOUS: Overload will reset itself
        LATCHING: Overload must be reset my command or manually at device

        :raises Pfa2FilterCommunicationError: When overload mode is unknown
        :return: Overload handling mode
        """
        response = self._com.query(
            _Pfa2FilterCommands.OVERLOAD_MODE.build_str(self._channel)  # type: ignore[attr-defined]
        )
        if response is None:
            err_msg = f"Overload mode for channel {self._channel} is not set or unknown"
            logger.error(err_msg)
            raise Pfa2FilterCommunicationError(err_msg)

        response = Pfa2FilterOverloadMode(response)
        logger.info(f"Overload mode for channel {self._channel} is '{response.name}'")
        return response

    @overload_mode.setter
    def overload_mode(self, value: Pfa2FilterOverloadMode | str) -> None:
        """
        Set the overload handling mode

        :param value: Value from Pfa2FilterOverloadMode
        :raises TypeError: When value is not a string
        :raises ValueError: When value is not in Pfa2FilterOverloadMode
        :raises Pfa2FilterCommunicationError: When overload mode could not be set
        """
        if not isinstance(value, str):
            raise TypeError
        if value not in Pfa2FilterOverloadMode:  # type: ignore[attr-defined]
            raise ValueError

        response = self._com.query(
            _Pfa2FilterCommands.OVERLOAD_MODE.build_str(self._channel, value)  # type: ignore[attr-defined]
        )
        if response is None:
            err_msg = f"Overload mode for channel {self._channel} could not be set"
            logger.error(err_msg)
            raise Pfa2FilterCommunicationError(err_msg)

        response = Pfa2FilterOverloadMode(response)
        logger.info(
            f"Setting overload mode to '{response.name}' for channel {self._channel}"
        )

    def overload_clear(self) -> bool:
        """
        Clear a overload

        :raises Pfa2FilterCommunicationError: When filter could not be reached
        :return: True if overload was cleared, False if not, None if in continuous mode
        """
        if self.overload_mode != Pfa2FilterOverloadMode.LATCHING:
            logger.info(
                f"Overload mode for channel {self._channel} is not "
                f"{Pfa2FilterOverloadMode.LATCHING.name}, no need to clear"  # type: ignore[attr-defined]
            )
            return not (self.input_overload or self.output_overload)

        response = self._com.query(
            _Pfa2FilterCommands.OVERLOAD_CLEAR.build_str(self._channel)  # type: ignore[attr-defined]
        )
        if response is None:
            err_msg = f"Overload for channel {self._channel} was not cleared"
            logger.error(err_msg)
            raise Pfa2FilterCommunicationError(err_msg)

        clear = not (self.input_overload or self.output_overload)
        logger.info(
            f"Overload for channel {self._channel} was {'' if clear else 'not '}"
            f"cleared successfully"
        )
        return clear

    @staticmethod
    def _clean_lpf_value(value: int) -> int:
        """
        Checks the value and rounds up to the next available value:
        - If value ≤ 2555, round up to the nearest multiple of 5 Hz
        - If 2555 < value ≤ 2750, use 2750
        - If value > 2750, round up to the nearest multiple of 250 Hz

        :param value: value to check
        :return: checked and rounded value
        """
        validate_number(
            "lpf_value",
            value,
            (LPF_FREQUENCY_MIN, LPF_FREQUENCY_MAX),
            int,
            logger=logger,
        )

        value_out = value

        if value <= 2555:
            value_out = ((value + 4) // 5) * 5  # Round up to nearest multiple of 5
        elif value <= 2750:
            value_out = 2750
        elif value <= 127750:
            value_out = (
                (value + 249) // 250
            ) * 250  # Round up to nearest multiple of 250

        if value_out != value:
            logger.info(f"The LPF frequency was changed to {value} Hz")
        return value_out

    @staticmethod
    def _clean_hpf_value(value: int) -> int:
        """
        Checks the value and rounds down to the next available hpf setting.
        - If value < 255 Hz round down to the nearest multiple of 1 Hz
        - If 255 Hz ≤ value < 300 Hz, use 255 Hz
        - If 300 Hz ≤ value < 12750 Hz, round down to the nearest multiple 50 Hz
        - If 12750 Hz ≤ value < 13000 Hz, use 12750 Hz
        - If 13000 Hz ≤ value, round down to the nearest multiple 500 Hz

        :param value: value to check
        :return: checked and rounded value
        """
        validate_number(
            "hpf_value",
            value,
            (HPF_FREQUENCY_MIN, HPF_FREQUENCY_MAX),
            int,
            logger=logger,
        )

        value_out = value

        if value < 255:
            pass
        elif value < 300:
            value_out = 255
        elif value < 12750:
            value_out = (value // 50) * 50  # Round down to nearest multiple of 50
        elif value < 13000:
            value_out = 12750
        elif value <= 127500:
            value_out = (value // 500) * 500  # Round down to nearest multiple of 500

        if value_out != value:
            logger.info(f"The HPF frequency was changed to {value} Hz")
        return value_out



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Facilities providing classes for handling configuration for communication protocols
and devices.
"""

import dataclasses
import json
import logging
from abc import ABC, abstractmethod
from collections.abc import Callable, Sequence
from importlib import import_module
from pathlib import Path

from typeguard import TypeCheckError, check_type  # type: ignore[import-not-found]
from typing_extensions import Self

logger = logging.getLogger(__name__)


def _has_default_value(f: dataclasses.Field):
    return (
        f.default is not dataclasses.MISSING
        or f.default_factory is not dataclasses.MISSING
    )


# Hooks of configdataclass
def _clean_values(_self):
    """
    Cleans and enforces configuration values. Does nothing by default, but may be
    overridden to add custom configuration value checks.
    """


_configclass_hooks = {
    "clean_values": _clean_values,
}


# Methods of configdataclass
def ___post_init__(self):  # noqa: N807
    self._check_types()
    self.clean_values()


def _force_value(self, fieldname, value):
    """
    Forces a value to a dataclass field despite the class being frozen.

    NOTE: you can define `post_force_value` method with same signature as this method
    to do extra processing after `value` has been forced on `fieldname`.

    :param fieldname: name of the field
    :param value: value to assign
    """
    object.__setattr__(self, fieldname, value)
    if hasattr(self, "post_force_value"):
        self.post_force_value(fieldname, value)


@classmethod  # type: ignore[misc]
def _keys(cls) -> Sequence[str]:
    """
    Returns a list of all configdataclass fields key-names.

    :return: a list of strings containing all keys.
    """
    return [f.name for f in dataclasses.fields(cls)]


@classmethod  # type: ignore[misc]
def _required_keys(cls) -> Sequence[str]:
    """
    Returns a list of all configdataclass fields, that have no default value assigned
    and need to be specified on instantiation.

    :return: a list of strings containing all required keys.
    """
    return [f.name for f in dataclasses.fields(cls) if not _has_default_value(f)]


@classmethod  # type: ignore[misc]
def _optional_defaults(cls) -> dict[str, object]:
    """
    Returns a list of all configdataclass fields, that have a default value assigned
    and may be optionally specified on instantiation.

    :return: a list of strings containing all optional keys.
    """
    return {
        f.name: f.default
        if f.default_factory is dataclasses.MISSING
        else f.default_factory()
        for f in dataclasses.fields(cls)
        if _has_default_value(f)
    }


def __check_types(self):
    mod = import_module(self.__module__)
    for field in dataclasses.fields(self):
        name = field.name
        value = getattr(self, name)
        type_ = field.type
        if isinstance(type_, str):  # `from __future__ import annotations` in use
            try:
                # built-in types
                type_ = eval(type_)  # noqa: S307
            except NameError:
                # no logging at this point, try-except controls program behaviour
                # module-level defined type
                type_ = getattr(mod, type_)
        try:
            check_type(value, type_)
        except TypeCheckError as exc:
            msg = (
                f"Type of field '{name}' is '{type(value)}' "
                f"and does not match '{type_}.'"
            )
            raise TypeError(msg) from exc


_configclass_methods = {
    "__post_init__": ___post_init__,
    "force_value": _force_value,
    "keys": _keys,
    "required_keys": _required_keys,
    "optional_defaults": _optional_defaults,
    "_check_types": __check_types,
}


def configdataclass(direct_decoration=None, frozen=True) -> Callable:
    """
    Decorator to make a class a configdataclass. Types in these dataclasses are
    enforced. Implement a function clean_values(self) to do additional checking on
    value ranges etc.

    It is possible to inherit from a configdataclass and re-decorate it with
    @configdataclass. In a subclass, default values can be added to existing fields.
    Note: adding additional non-default fields is prone to errors, since the order
    has to be respected through the whole chain (first non-default fields, only then
    default-fields).

    :param frozen: defaults to True. False allows to later change configuration values.
        Attention: if configdataclass is not frozen and a value is changed, typing is
        not enforced anymore!
    """

    def decorator(cls):
        for name, method in _configclass_methods.items():
            if name in cls.__dict__:
                msg = f"configdataclass {cls.__name__} cannot define {name} method"
                raise AttributeError(msg)
            setattr(cls, name, method)
        for name, hook in _configclass_hooks.items():
            if not hasattr(cls, name):
                setattr(cls, name, hook)
        if not hasattr(cls, "is_configdataclass"):
            cls.is_configdataclass = True

        return dataclasses.dataclass(cls, frozen=frozen)

    if direct_decoration:
        return decorator(direct_decoration)

    return decorator


class ConfigurationMixin(ABC):
    """
    Mixin providing configuration to a class.
    """

    # omitting type hint of `configuration` on purpose, because type hinting
    # configdataclass is not possible. Union[dict[str, object], object] resolves to
    # object.
    def __init__(self, configuration) -> None:
        """
        Constructor for the configuration mixin.

        :param configuration: is the configuration provided either as:
        *   a dict with string keys and values, then the default config dataclass
            will be used
        *   a configdataclass object
        *   None, then the config_cls() with no parameters is instantiated
        """

        if not configuration:
            configuration = {}

        if hasattr(configuration, "is_configdataclass"):
            self._configuration = configuration
        elif isinstance(configuration, dict):
            default_configdataclass = self.config_cls()
            if not hasattr(default_configdataclass, "is_configdataclass"):
                msg = (
                    "Default configdataclass is not a configdataclass. Is"
                    "the decorator `@configdataclass` applied?"
                )
                raise TypeError(msg)
            self._configuration = default_configdataclass(**configuration)
        else:
            msg = "configuration is not a dictionary or configdataclass."
            raise TypeError(msg)

    @staticmethod
    @abstractmethod
    def config_cls():
        """
        Return the default configdataclass class.

        :return: a reference to the default configdataclass class
        """

    @property
    def config(self):  # noqa: ANN201
        """
        ConfigDataclass property.

        :return: the configuration
        """

        return self._configuration

    @classmethod
    def from_json(cls, filename: str) -> Self:
        """
        Instantiate communication protocol using configuration from a JSON file.

        :param filename: Path and filename to the JSON configuration
        """

        configuration = cls._configuration_load_json(filename)
        return cls(configuration)

    def configuration_save_json(self, path: str) -> None:
        """
        Save current configuration as JSON file.

        :param path: path to the JSON file.
        """

        self._configuration_save_json(dataclasses.asdict(self._configuration), path)

    @staticmethod
    def _configuration_load_json(path: str) -> dict[str, object]:
        """
        Load configuration from JSON file and return dict. This method is only used
        during construction, if not directly a configuration is given but rather a
        path to a JSON config file.

        :param path: Path to the JSON configuration file.
        :return: Dictionary containing the parameters read from the JSON file.
        """

        with Path(path).open() as fp:
            return json.load(fp)

    @staticmethod
    def _configuration_save_json(configuration: dict[str, object], path: str) -> None:
        """
        Store a configuration dict to a JSON file.

        :param configuration: configuration dictionary
        :param path: path to the JSON file.
        """

        with Path(path).open("w") as fp:
            json.dump(configuration, fp, indent=4)


@configdataclass
class EmptyConfig:
    """
    Empty configuration dataclass.
    """


class ConfigurationValueWarning(UserWarning):
    """
    User warnings category for values of `@configdataclass` fields.
    """



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Tests for the base class CommunicationProtocol.
"""

import pytest

from hvl_ccb.comm import AsyncCommunicationProtocolConfig, NullCommunicationProtocol
from hvl_ccb.configuration import EmptyConfig


@pytest.fixture(scope="module")
def com_config():
    return {
        "terminator": b"",
        "encoding": "ascii",
        "encoding_error_handling": "strict",
        "wait_sec_read_text_nonempty": 0.01,
        "default_n_attempts_read_text_nonempty": 5,
    }


def test_com_config(com_config) -> None:
    config = AsyncCommunicationProtocolConfig(**com_config)
    for key, value in com_config.items():
        assert getattr(config, key) == value


@pytest.mark.parametrize(
    "wrong_config_dict",
    [
        {"encoding": "enigma"},
        {"encoding_error_handling": "miracle"},
        {"wait_sec_read_text_nonempty": 0},
        {"wait_sec_read_text_nonempty": -1},
        {"default_n_attempts_read_text_nonempty": 0},
        {"default_n_attempts_read_text_nonempty": -1},
    ],
)
def test_invalid_config_dict(com_config, wrong_config_dict) -> None:
    invalid_config = dict(com_config)
    invalid_config.update(wrong_config_dict)
    with pytest.raises(ValueError):
        AsyncCommunicationProtocolConfig(**invalid_config)


def test_instantiation() -> None:
    for arg in (EmptyConfig(), {}, None):
        with NullCommunicationProtocol(arg) as com:
            assert com is not None
            assert isinstance(com.config, EmptyConfig)

    with pytest.raises(TypeError):
        NullCommunicationProtocol({"extra_key": 0})



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Fluke 8845A multimeter implementation using TCP communication
"""

from .base import (  # noqa: F401
    Fluke8845a,
    Fluke8845aCommunication,
    Fluke8845aCommunicationConfig,
    Fluke8845aConfig,
)
from .constants import Fluke8845aError, MeasurementFunction, TriggerSource  # noqa: F401
from .ranges import (  # noqa: F401
    ACCurrentRange,
    ACVoltageRange,
    ApertureRange,
    DCCurrentRange,
    DCVoltageRange,
    FilterRange,
    ResistanceRange,
)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
A LabJack T-series devices wrapper around the LabJack's LJM Library; see
https://labjack.com/ljm .
The wrapper was originally developed and tested for a LabJack T7-PRO device.

Extra installation
~~~~~~~~~~~~~~~~~~

To use this LabJack T-series devices wrapper:

1. install the :code:`hvl_ccb` package with a :code:`labjack` extra feature::

        $ pip install "hvl_ccb[labjack]"

   this will install the Python bindings for the library.

2. install the library - follow instruction in
   https://labjack.com/support/software/installers/ljm .

"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, cast

from aenum import Enum, IntEnum

from hvl_ccb._dev import labjack
from hvl_ccb.comm.labjack_ljm import LJMCommunication
from hvl_ccb.dev import SingleCommDevice
from hvl_ccb.dev.base import DeviceError
from hvl_ccb.utils.enum import NameEnum, StrEnumBase
from hvl_ccb.utils.validation import validate_bool, validate_number

if TYPE_CHECKING:
    from collections.abc import Sequence

    from hvl_ccb.utils.typing import Number

logger = logging.getLogger(__name__)


class LabJackError(DeviceError):
    """
    General Error for the LabJack device.
    """


class LabJackIdentifierDIOError(LabJackError):
    """
    Error indicating a wrong DIO identifier
    """


class LabJack(SingleCommDevice):
    """
    LabJack Device.

    This class is tested with a LabJack T7-Pro and should also work with T4 and T7
    devices communicating through the LJM Library. Other or older hardware versions and
    variants of LabJack devices are not supported.
    """

    DeviceType = labjack.DeviceType
    """
    LabJack device types.
    """

    def __init__(self, com, dev_config=None) -> None:
        """
        Constructor for a LabJack Device.

        :param com: Communication protocol object of type
            LJMCommunication. If a configuration (dict or configdataclass) is given,
            a new communication protocol object will be instantiated.
        :param dev_config: There is no device configuration for LabJack yet.
        """
        super().__init__(com, dev_config)

        # cached device type
        self._device_type: labjack.DeviceType | None = None

        # clock configuration for pulse out feature
        self.configured_pulse_addresses: list[str | labjack.TSeriesDIOChannel] = []

    @staticmethod
    def default_com_cls():
        return LJMCommunication

    def start(self) -> None:
        """
        Start the Device.
        """

        logger.info(f"Starting device {self!s}")
        super().start()

    def stop(self) -> None:
        """
        Stop the Device.
        """

        logger.info(f"Stopping device {self!s}")
        super().stop()

    def _read_float(self, *names: str) -> float | Sequence[float]:
        """
        Read a numeric value.

        :param name: name to read via communication protocol
        :return: read numeric value
        """
        return self.com.read_name(*names, return_num_type=float)

    def _read_int(self, *names: str) -> int | Sequence[int]:
        """
        Read an integer value.

        :param name: name to read via communication protocol
        :return: read integer value
        """
        return self.com.read_name(*names, return_num_type=int)

    def get_serial_number(self) -> int:
        """
        Returns the serial number of the connected LabJack.

        :return: Serial number.
        """

        return cast("int", self._read_int("SERIAL_NUMBER"))

    def get_sbus_temp(self, number: int) -> float:
        """
        Read the temperature value from a serial SBUS sensor.

        :param number: port number (0..22)
        :return: temperature in Kelvin
        """

        return cast("float", self._read_float(f"SBUS{number}_TEMP"))

    def get_sbus_rh(self, number: int) -> float:
        """
        Read the relative humidity value from a serial SBUS sensor.

        :param number: port number (0..22)
        :return: relative humidity in %RH
        """

        return cast("float", self._read_float(f"SBUS{number}_RH"))

    class AInRange(StrEnumBase, init="value_str"):  # type: ignore[call-arg]
        TEN = "10"
        ONE = "1"
        ONE_TENTH = "0.1"
        ONE_HUNDREDTH = "0.01"

        def __str__(self) -> str:
            return self.value_str

        @property
        def value(self) -> float:
            return float(self.value_str)

    def get_ain(self, *channels: int) -> float | Sequence[float]:
        """
        Read currently measured value (voltage, resistance, ...) from one or more
        of analog inputs.

        :param channels: AIN number or numbers (0..254)
        :return: the read value (voltage, resistance, ...) as `float`or `tuple` of
            them in case multiple channels given
        """
        ch_str = [f"AIN{ch}" for ch in channels]
        return self._read_float(*ch_str)

    def set_analog_output(self, channel: int, value: Number) -> None:
        """
        Set the voltage of a analog output port

        :param channel: DAC channel number 1/0
        :param value: The output voltage value 0-5 Volts int/float
        """
        validate_number("DAC channel number", channel, (0, 1), int, logger)
        validate_number("DAC output voltage", value, (0.0, 5.0), logger=logger)
        self.com.write_name(f"DAC{channel}", value)

    def set_ain_range(self, channel: int, vrange: Number | AInRange) -> None:
        """
        Set the range of an analog input port.

        :param channel: is the AIN number (0..254)
        :param vrange: is the voltage range to be set
        """
        vrange = self.AInRange(str(vrange))
        self.com.write_name(f"AIN{channel}_RANGE", vrange.value)

    def set_ain_resolution(self, channel: int, resolution: int) -> None:
        """
        Set the resolution index of an analog input port.

        :param channel: is the AIN number (0..254)
        :param resolution: is the resolution index within
            0...`get_product_type().ain_max_resolution` range; 0 will set the
            resolution index to default value.
        """

        ain_max_resolution = self.get_product_type().ain_max_resolution
        if resolution not in range(ain_max_resolution + 1):
            msg = f"Not supported resolution index: {resolution}"
            raise LabJackError(msg)

        self.com.write_name(f"AIN{channel}_RESOLUTION_INDEX", resolution)

    def set_ain_differential(self, pos_channel: int, differential: bool) -> None:
        """
        Sets an analog input to differential mode or not.
        T7-specific: For base differential channels, positive must be even channel
        from 0-12 and negative must be positive+1. For extended channels 16-127,
        see Mux80 datasheet.

        :param pos_channel: is the AIN number (0..12)
        :param differential: True or False
        :raises LabJackError: if parameters are unsupported
        """

        if pos_channel not in range(13):
            msg = f"Not supported pos_channel: {pos_channel}"
            raise LabJackError(msg)

        if pos_channel % 2 != 0:
            msg = (
                "AIN pos_channel for positive part of differential pair"
                f" must be even: {pos_channel}"
            )
            raise LabJackError(msg)

        neg_channel = pos_channel + 1

        self.com.write_name(
            f"AIN{pos_channel}_NEGATIVE_CH", neg_channel if differential else 199
        )

    class ThermocoupleType(NameEnum, init="ef_index"):  # type: ignore[call-arg]
        """
        Thermocouple type; NONE means disable thermocouple mode.
        """

        NONE = 0
        E = 20
        J = 21
        K = 22
        R = 23
        T = 24
        S = 25
        C = 30
        PT100 = 40
        PT500 = 41
        PT1000 = 42

    class CjcType(NameEnum, init="slope offset"):  # type: ignore[call-arg]
        """
        CJC slope and offset
        """

        internal = 1, 0
        lm34 = 55.56, 255.37

    class TemperatureUnit(NameEnum, init="ef_config_a"):  # type: ignore[call-arg]
        """
        Temperature unit (to be returned)
        """

        K = 0
        C = 1
        F = 2

    def set_ain_thermocouple(
        self,
        pos_channel: int,
        thermocouple: None | str | ThermocoupleType,
        cjc_address: int = 60050,
        cjc_type: str | CjcType = (CjcType.internal),  # type: ignore[assignment]
        vrange: Number | AInRange = (AInRange.ONE_HUNDREDTH),  # type: ignore[assignment]
        resolution: int = 10,
        unit: str | TemperatureUnit = (TemperatureUnit.K),  # type: ignore[assignment]
    ) -> None:
        """
        Set the analog input channel to thermocouple mode.

        :param pos_channel: is the analog input channel of the positive part of the
            differential pair
        :param thermocouple: None to disable thermocouple mode, or string specifying
            the thermocouple type
        :param cjc_address: modbus register address to read the CJC temperature
        :param cjc_type: determines cjc slope and offset, 'internal' or 'lm34'
        :param vrange: measurement voltage range
        :param resolution: resolution index (T7-Pro: 0-12)
        :param unit: is the temperature unit to be returned ('K', 'C' or 'F')
        :raises LabJackError: if parameters are unsupported
        """

        if thermocouple is None:
            thermocouple = self.ThermocoupleType.NONE  # type: ignore[assignment]

        thermocouple = self.ThermocoupleType(thermocouple)

        # validate separately from `set_ain_range` to fail before any write happens
        # (in `set_ain_differential` first)
        vrange = self.AInRange(str(vrange))

        unit = self.TemperatureUnit(unit)

        cjc_type = self.CjcType(cjc_type)

        self.set_ain_differential(pos_channel=pos_channel, differential=True)
        self.set_ain_range(pos_channel, vrange)
        self.set_ain_resolution(pos_channel, resolution)
        self.set_ain_range(pos_channel + 1, vrange)
        self.set_ain_resolution(pos_channel + 1, resolution)

        # specify thermocouple mode
        self.com.write_name(f"AIN{pos_channel}_EF_INDEX", thermocouple.ef_index)

        # specify the units for AIN#_EF_READ_A and AIN#_EF_READ_C (0 = K, 1 = C, 2 = F)
        self.com.write_name(f"AIN{pos_channel}_EF_CONFIG_A", unit.ef_config_a)

        # specify modbus address for cold junction reading CJC
        self.com.write_name(f"AIN{pos_channel}_EF_CONFIG_B", cjc_address)

        # set slope for the CJC reading, typically 1
        self.com.write_name(f"AIN{pos_channel}_EF_CONFIG_D", cjc_type.slope)

        # set the offset for the CJC reading, typically 0
        self.com.write_name(f"AIN{pos_channel}_EF_CONFIG_E", cjc_type.offset)

    def read_thermocouple(self, pos_channel: int) -> float:
        """
        Read the temperature of a connected thermocouple.

        :param pos_channel: is the AIN number of the positive pin
        :return: temperature in specified unit
        """

        return round(cast("float", self._read_float(f"AIN{pos_channel}_EF_READ_A")), 2)

    class DIOStatus(IntEnum):
        """
        State of a digital I/O channel.
        """

        LOW = 0
        HIGH = 1

    def set_digital_output(self, address: str, state: int | DIOStatus) -> None:
        """
        Set the value of a digital output.

        :param address: name of the output -> `'FIO0'`
        :param state: state of the output -> `DIOStatus` instance or corresponding `int`
            value
        """
        dt = self.get_product_type()
        if address not in (dt.dio):
            raise LabJackIdentifierDIOError
        state = self.DIOStatus(state)
        self.com.write_name(address, state)

    DIOChannel = labjack.TSeriesDIOChannel

    def get_digital_input(
        self, address: str | labjack.TSeriesDIOChannel
    ) -> LabJack.DIOStatus:
        """
        Get the value of a digital input.

        allowed names for T7 (Pro): FIO0 - FIO7, EIO0 - EIO 7, CIO0- CIO3, MIO0 - MIO2
        :param address: name of the output -> 'FIO0'
        :return: HIGH when `address` DIO is high, and LOW when `address` DIO is low
        """
        if not isinstance(address, self.DIOChannel):
            address = self.DIOChannel(address)
        dt = self.get_product_type()
        if address not in (dt.dio):
            dt_name = dt.name
            msg = (
                f"DIO {address.name} is not available for this device type: {dt_name}."
            )
            raise LabJackIdentifierDIOError(msg)
        try:
            ret = self._read_int(address.name)
            return self.DIOStatus(ret)
        except ValueError as exc:
            msg = f"Expected 0 or 1 return value, got {ret}."
            logger.exception(msg)
            raise LabJackIdentifierDIOError(msg) from exc

    class CalMicroAmpere(Enum, init="value current_source_query"):  # type: ignore[call-arg]
        """
        Pre-defined microampere (uA) values for calibration current source query.
        """

        TEN = "10uA", "CURRENT_SOURCE_10UA_CAL_VALUE"
        TWO_HUNDRED = "200uA", "CURRENT_SOURCE_200UA_CAL_VALUE"

    def get_cal_current_source(self, name: str | CalMicroAmpere) -> float:
        """
        This function will return the calibration of the chosen current source,
        this ist not a measurement!

        The value was stored during fabrication.

        :param name: '200uA' or '10uA' current source
        :return: calibration of the chosen current source in ampere
        """
        if not isinstance(name, self.CalMicroAmpere):
            name = self.CalMicroAmpere(name)
        return cast("float", self._read_float(name.current_source_query))

    def get_product_id(self) -> int:
        """
        This function returns the product ID reported by the connected device.

        Attention: returns `7` for both T7 and T7-Pro devices!

        :return: integer product ID of the device
        """
        return cast("int", self._read_int("PRODUCT_ID"))

    def get_product_type(self, force_query_id: bool = False) -> labjack.DeviceType:
        """
        This function will return the device type based on reported device type and
        in case of unambiguity based on configuration of device's communication
        protocol (e.g. for "T7" and  "T7_PRO" devices), or, if not available first
        matching.


        :param force_query_id: boolean flag to force `get_product_id` query to device
            instead of using cached device type from previous queries.
        :return: `DeviceType` instance
        :raises LabJackIdentifierDIOError: when read Product ID is unknown
        """
        if force_query_id or not self._device_type:
            try:
                device_type_or_list = self.DeviceType.get_by_p_id(self.get_product_id())
            except ValueError as e:
                msg = "Error: Unknown Product ID"
                logger.exception(msg, exc_info=e)
                raise LabJackIdentifierDIOError(msg) from e
            if isinstance(device_type_or_list, self.DeviceType):
                device_type = device_type_or_list
            else:  # isinstance(device_type_or_list, list):
                device_type_list: list[labjack.DeviceType] = device_type_or_list
                # can be None in case a non-default com or its config was used
                conf_device_type = getattr(self.com.config, "device_type", None)
                if conf_device_type:
                    if conf_device_type not in device_type_list:
                        msg = (
                            f"Configured devices type {conf_device_type!s} does not "
                            "match any of the unambiguously reported device types: "
                            f"{','.join(str(dt) for dt in device_type_list)}."
                        )
                        raise LabJackIdentifierDIOError(msg)
                    device_type = conf_device_type
                else:
                    device_type = device_type_list[0]
            self._device_type = device_type
        return self._device_type

    def get_product_name(self, force_query_id=False) -> str:
        """
        This function will return the product name based on product ID reported by
        the device.

        Attention: returns "T7" for both T7 and T7-Pro devices!

        :param force_query_id: boolean flag to force `get_product_id` query to device
            instead of using cached device type from previous queries.
        :return: device name string, compatible with `LabJack.DeviceType`
        """
        return self.get_product_type(force_query_id=force_query_id).name

    def set_ain_resistance(
        self, channel: int, vrange: Number | AInRange, resolution: int
    ) -> None:
        """
        Set the specified channel to resistance mode. It utilized the 200uA current
        source of the LabJack.

        :param channel: channel that should measure the resistance
        :param vrange: voltage range of the channel
        :param resolution: resolution index of the channel T4: 0-5, T7: 0-8, T7-Pro 0-12
        """
        self.set_ain_range(channel, vrange)
        self.set_ain_resolution(channel, resolution)

        # resistance mode
        self.com.write_name(f"AIN{channel}_EF_INDEX", 4)
        # excitation with 200uA current source
        self.com.write_name(f"AIN{channel}_EF_CONFIG_B", 0)

    def read_resistance(self, channel: int) -> float:
        """
        Read resistance from specified channel.

        :param channel: channel with resistor
        :return: resistance value with 2 decimal places
        """
        return round(cast("float", self._read_float(f"AIN{channel}_EF_READ_A")), 2)

    class ClockFrequency(IntEnum):
        """
        Available clock frequencies, in Hz
        """

        MAXIMUM = 80_000_000
        FORTY_MHZ = 40_000_000
        TWENTY_MHZ = 20_000_000
        TEN_MHZ = 10_000_000
        FIVE_MHZ = 5_000_000
        TWENTY_FIVE_HUNDRED_KHZ = 2_500_000
        TWELVE_HUNDRED_FIFTY_KHZ = 1_250_000
        MINIMUM = 312_500

    class BitLimit(IntEnum):
        """
        Maximum integer values for clock settings
        """

        THIRTY_TWO_BIT = 2**32 - 1

    @property
    def _clock_config(self) -> dict[str, int]:
        clock_config = {
            "divisor": self.com.read_name("DIO_EF_CLOCK0_DIVISOR"),
            "roll_value": self.com.read_name("DIO_EF_CLOCK0_ROLL_VALUE"),
        }
        # adjust zeroes to corresponding real values for frequency/period conversion:
        if clock_config["divisor"] == 0:
            clock_config["divisor"] = 1
        if clock_config["roll_value"] == 0:
            clock_config["roll_value"] = self.BitLimit.THIRTY_TWO_BIT

        return clock_config

    @_clock_config.setter
    def _clock_config(self, clock_config: dict[str, int]):
        self.com.write_name("DIO_EF_CLOCK0_DIVISOR", clock_config["divisor"])
        self.com.write_name("DIO_EF_CLOCK0_ROLL_VALUE", clock_config["roll_value"])

    def enable_clock(self, clock_enabled: bool) -> None:
        """
        Enable/disable LabJack clock to configure or send pulses.
        :param clock_enabled: True -> enable, False -> disable.
        :raises TypeError: if clock_enabled is not of type bool
        """
        validate_bool("clock enabled", clock_enabled, logger=logger)
        self.com.write_name("DIO_EF_CLOCK0_ENABLE", int(clock_enabled))

    def get_clock(self) -> dict[str, Number]:
        """
        Return clock settings read from LabJack.
        """
        divisor, roll_value = self._clock_config.values()
        clock_frequency = self.ClockFrequency.MAXIMUM / divisor
        clock_period = roll_value / clock_frequency
        return {
            "clock_frequency": clock_frequency,
            "clock_period": clock_period,
        }

    def set_clock(
        self,
        clock_frequency: Number | ClockFrequency = 10_000_000,
        clock_period: Number = 1,
    ) -> None:
        """
        Configure LabJack clock for pulse out feature.
        :param clock_frequency: clock frequency in Hz; default 10 MHz for base 10.
        :raises ValueError: if clock_frequency is not allowed (see ClockFrequency).
        :param clock_period: clock roll time in seconds; default 1s, 0 for max.
        :raises ValueError: if clock_period exceeds the 32bit tick limit.
        Clock period determines pulse spacing when using multi-pulse settings.
        Ensure period exceeds maximum intended pulse end time.
        """
        validate_number(
            "clock frequency",
            clock_frequency,
            limits=(self.ClockFrequency.MINIMUM, self.ClockFrequency.MAXIMUM),
            logger=logger,
        )
        clock_frequency = self.ClockFrequency(clock_frequency)
        divisor = int(self.ClockFrequency.MAXIMUM / clock_frequency)
        validate_number(
            "clock period",
            clock_period,
            limits=(0, self.BitLimit.THIRTY_TWO_BIT / clock_frequency),
            logger=logger,
        )
        clock_period_ticks = int(clock_period * clock_frequency)

        self.enable_clock(False)
        self._clock_config = {"divisor": divisor, "roll_value": clock_period_ticks}
        # adjust zero input to real period for logging.
        if clock_period == 0:
            clock_period = self.BitLimit.THIRTY_TWO_BIT / clock_frequency
        logger.info(f"Clock frequency is set to {clock_frequency:_} Hz")
        logger.info(f"Clock period is set to {clock_period} s")

    def config_high_pulse(
        self,
        address: str | labjack.TSeriesDIOChannel,
        t_start: Number,
        t_width: Number,
        n_pulses: int = 1,
    ) -> None:
        """
        Configures one FIO channel to send a timed HIGH pulse.
        Configure multiple channels to send pulses with relative timing accuracy.
        Times have a maximum resolution of 1e-7 seconds @ 10 MHz.
        :param address: FIO channel: [T7] FIO0;2;3;4;5. [T4] FIO6;7.
        :raises LabJackError if address is not supported.
        :param t_start: pulse start time in seconds.
        :raises ValueError: if t_start is negative or would exceed the clock period.
        :param t_width: duration of high pulse, in seconds.
        :raises ValueError: if t_width is negative or would exceed the clock period.
        :param n_pulses: number of pulses to be sent; single pulse default.
        :raises TypeError if n_pulses is not of type int.
        :raises Value Error if n_pulses is negative or exceeds the 32bit limit.
        """
        if not isinstance(address, self.DIOChannel):
            address = self.DIOChannel(address)

        device_type = self.get_product_type()
        if address not in device_type.pulse_out_addr:
            msg = f"{address} does not support pulse feature"
            logger.error(msg)
            raise LabJackError(msg)

        clock_settings = self.get_clock()
        clock_frequency = clock_settings["clock_frequency"]
        clock_period = clock_settings["clock_period"]

        validate_number("pulse start", t_start, limits=(0, clock_period), logger=logger)
        validate_number(
            "pulse width", t_width, limits=(0, clock_period - t_start), logger=logger
        )
        validate_number(
            "n pulses",
            n_pulses,
            limits=(0, self.BitLimit.THIRTY_TWO_BIT),
            number_type=int,
            logger=logger,
        )

        start_tick = int(t_start * clock_frequency) + 1
        end_tick = int((t_start + t_width) * clock_frequency) + 1
        addr = address.address

        self.enable_clock(False)
        self.com.write_names(
            {
                f"DIO{addr}": 0,
                f"DIO{addr}_EF_ENABLE": 0,
                f"DIO{addr}_EF_OPTIONS": 0,
                f"DIO{addr}_EF_INDEX": 2,
                f"DIO{addr}_EF_CONFIG_B": start_tick,
                f"DIO{addr}_EF_CONFIG_A": end_tick,
                f"DIO{addr}_EF_CONFIG_C": n_pulses,
            }
        )

        if address not in self.configured_pulse_addresses:
            self.configured_pulse_addresses.append(address)
        logger.info(f"Pulse configured for {address}")

    def send_pulses(self, *addresses: str | labjack.TSeriesDIOChannel) -> None:
        """
        Sends pre-configured pulses for specified addresses.
        :param addresses: tuple of FIO addresses
        :raises LabJackError if an address has not been configured.
        """
        address_list = []
        for address in addresses:
            if not isinstance(address, self.DIOChannel):
                address = self.DIOChannel(address)
            if address not in self.configured_pulse_addresses:
                msg = f"No pulse configured for {address}"
                logger.error(msg)
                raise LabJackError(msg)
            address_list.append(address.address)

        self.enable_clock(False)
        for addr in address_list:
            self.com.write_name(f"DIO{addr}_EF_ENABLE", 1)

        self.enable_clock(True)
        logger.info(f"Pulses sent on {addresses}")

    def disable_pulses(
        self, *addresses: str | labjack.TSeriesDIOChannel | None
    ) -> None:
        """
        Disable previously configured pulse channels.
        :param addresses: tuple of FIO addresses.
        All channels disabled if no argument is passed.
        """
        self.enable_clock(False)
        if len(addresses) == 0:
            addresses = tuple(self.configured_pulse_addresses)

        for address in addresses:
            if not isinstance(address, self.DIOChannel):
                address = self.DIOChannel(address)
            if address not in self.configured_pulse_addresses:
                logger.warning(f"No pulse configured for {address}")
                continue
            self.com.write_name(f"DIO{address.address}_EF_ENABLE", 0)
            self.configured_pulse_addresses.remove(address.name)
            logger.info(f"Pulse disabled for {address}")



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#

import contextlib
import logging
import re
from collections.abc import Generator
from time import sleep
from typing import Any

from hvl_ccb.configuration import configdataclass
from hvl_ccb.dev.base import SingleCommDevice
from hvl_ccb.utils.validation import validate_bool, validate_number

from .base import KA3000PError, KA3000PStatus, _gen_status
from .comm import KA3000PCommunication

logger = logging.getLogger(__name__)


@configdataclass
class KA3000PConfig:
    """Device configuration class for a KA3000P power supply"""

    time_to_wait_before_save_settings: float = 0.1


class KA3000P(SingleCommDevice):
    """Class to controll a KA3000P power supply"""

    com: KA3000PCommunication
    config: KA3000PConfig

    @staticmethod
    def default_com_cls():
        return KA3000PCommunication

    @staticmethod
    def config_cls():
        return KA3000PConfig

    def __init__(self, com, dev_config=None) -> None:
        super().__init__(com, dev_config)
        self._brand: str = ""
        self._comm_version: str = ""
        self._serial_number: str = ""
        # maximum output voltage of the hardware (unit V)
        self._max_voltage_hardware: float = 0
        # maximum output current of the hardware (unit A)
        self._max_current_hardware: float = 0

    def start(self) -> None:
        super().start()
        try:
            self.idenify_device()
        except KA3000PError as exc:
            self.stop()
            msg = (
                "Error during identifying the device. "
                "Either the device is unknown or the connection is not stable."
            )
            logger.exception(msg)
            raise KA3000PError(msg) from exc

    def idenify_device(self) -> None:
        identification: str = self.com.query_not_none("*IDN?")
        logger.debug(f'Identified device: "{identification}"')
        # Result: KORAD KA3005P V5.9 SN:02738566
        regex_model = (
            r"(?P<brand>[A-Z]+) "  # KORAD
            r"KA(?P<voltage>\d{2})(?P<current>\d{2})P "  # KA3005P
            r"V(?P<version>[.0-9]+) "  # V5.9
            r"SN:(?P<sn>[0-9]+)"  # SN:02738566
        )
        m_obj: re.Match[str] | None = re.match(regex_model, identification)
        if m_obj is None:
            msg = f"Cannot identify device with identification: {identification}"
            logger.error(msg)
            raise KA3000PError(msg)
        device_values: dict[str, str] = m_obj.groupdict()
        self._brand = device_values["brand"]
        self._max_voltage_hardware = float(device_values["voltage"])
        self._max_current_hardware = float(device_values["current"])
        self._comm_version = f"V{device_values['version']}"
        self._serial_number = device_values["sn"]

    # Static information
    @property
    def brand(self) -> str:
        """Brand of the connected device

        :return: Band
        """
        return self._brand

    @property
    def comm_version(self) -> str:
        """Version of the communication protocol

        :return: Version number
        """
        return self._comm_version

    @property
    def serial_number(self) -> str:
        """Serial number of the connected device

        :return: Serial number
        """
        return self._serial_number

    @property
    def status(self) -> KA3000PStatus:
        with self.com.access_lock:
            self.com.write("STATUS?")
            stb_qry = self.com.read_bytes()

        # A somehow "manual" strip of a single byte, as the result can be a byte which
        # could be interpreted as whitespace
        if len(stb_qry) == 0:
            msg = "Queried status and did not get an answer"
            logger.error(msg)
            raise KA3000PError(msg)
        if len(stb_qry) > 1 and len(stb_qry[1:].strip()) != 0:
            msg = f"Queried status and answer was not a single byte ({stb_qry!r})"
            logger.error(msg)
            raise KA3000PError(msg)
        return _gen_status(stb_qry[0])

    # Voltage and current values
    @property
    def voltage(self) -> float:
        """Return the measured output voltage in V"""
        value = float(self.com.query_not_none("VOUT1?"))
        logger.debug(f"Voltage at Output 1 of KA3000P: {value} V")
        return value

    @voltage.setter
    def voltage(self, value: float) -> None:
        """Set the output voltage"""
        validate_number("voltage", value, (0, self.max_voltage), logger=logger)
        value_str = f"VSET1:{value:.2f}"
        logger.debug(f'Set output voltage ({value} V) with: "{value_str}"')
        self.com.write(value_str)

    @property
    def current(self) -> float:
        """Return the measured output current in A"""
        value = float(self.com.query_not_none("IOUT1?"))
        logger.debug(f"Current at Output 1 of KA3000P: {value} A")
        return value

    @current.setter
    def current(self, value: float) -> None:
        """Set the output current"""
        validate_number("current", value, (0, self.max_current), logger=logger)
        value_str = f"ISET1:{value:.3f}"
        logger.debug(f'Set output current ({value} A) with: "{value_str}"')
        self.com.write(value_str)

    @property
    def set_voltage(self) -> float:
        """Return the set voltage (may differ from actual value) in V"""
        value = float(self.com.query_not_none("VSET1?"))
        logger.debug(f"Set Voltage at Output 1 of KA3000P: {value} V")
        return value

    @set_voltage.setter
    def set_voltage(self, value: float) -> None:
        """Set the output voltage"""
        self.voltage = value

    @property
    def set_current(self) -> float:
        """Return the set current (may differ from actual value) in A"""
        value = float(self.com.query_not_none("ISET1?"))
        logger.debug(f"Set Current at Output 1 of KA3000P: {value} A")
        return value

    @set_current.setter
    def set_current(self, value: float) -> None:
        """Set the output current"""
        self.current = value

    @property
    def max_voltage(self) -> float:
        """Maximal output voltage of the hardware in V,
        but user can reset to a lower value"""
        return self._max_voltage_hardware

    @property
    def max_current(self) -> float:
        """Maximal output current of the hardware in A,
        but user can reset to a lower value"""
        return self._max_current_hardware

    # Output stage
    @property
    def output(self) -> bool | None:
        """State of the high voltage output"""
        return self.status.output

    @output.setter
    def output(self, value: bool) -> None:
        """
        Activates the output of the source

        :param value: `True` for activation, `False` for deactivation
        """
        validate_bool("output", value, logger=logger)
        value_str = f"OUT{1 if value else 0}"
        logger.debug(f'Set output ({"ON" if value else "OFF"}) with: "{value_str}"')
        self.com.write(value_str)

    # Beep
    @property
    def beep(self) -> bool:
        """State of the beep"""
        return self.status.beep

    @beep.setter
    def beep(self, value: bool) -> None:
        """
        Activates the beep of the source

        :param value: `True` for activation, `False` for deactivation
        """
        validate_bool("beep", value, logger=logger)
        value_str = f"BEEP{1 if value else 0}"
        logger.debug(f'Set beep ({"ON" if value else "OFF"}) with: "{value_str}"')
        self.com.write(value_str)

    # Lock Screen
    @property
    def lock(self) -> bool:
        """State of the button lock"""
        msg = "The state of the button lock cannot be read. It is write-only..."
        logger.error(msg)
        raise NotImplementedError(msg)

    @lock.setter
    def lock(self, value: bool) -> None:
        """Lock or unlock the hardware buttons at the front panel

        :param value: `True` to lock the buttons, `False` to enable the manuel buttons
        """
        validate_bool("lock", value, logger=logger)
        value_str = f"LOCK{1 if value else 0}"
        logger.debug(f'Set lock ({"ON" if value else "OFF"}) with: "{value_str}"')
        self.com.write(value_str)

    # Controlled Voltage or Current
    @property
    def controlled_voltage(self) -> bool:
        """If the power supply operates at voltage limiting mode.
        Opposite to `controlled_current`.

        :return: Output is voltage limited
        """

        return self.status.ch1_cv

    @property
    def controlled_current(self) -> bool:
        """If the power supply operates at current limiting mode.
        Opposite to `controlled_voltage`.

        :return: Output is current limited
        """

        return not self.status.ch1_cv

    # Save and Recall settings (unfortunately, also write-only)
    @property
    def save(self) -> int:
        """Save a setting"""
        msg = "The save the settings cannot be read. It is write-only..."
        logger.error(msg)
        raise NotImplementedError(msg)

    @save.setter
    def save(self, value: int) -> None:
        """Save the panel settings to a given slot. Available slot: 1...5

        :param value: `int` between 1 and 5 to select the slot
        """
        validate_number("save", value, (1, 5), number_type=int, logger=logger)
        value_str = f"SAV{value}"
        logger.debug(f'Save settings to slot {value} with: "{value_str}"')
        self.com.write(value_str)

    @property
    def recall(self) -> int:
        """Recall a setting"""
        msg = "The recall a setting cannot be read. It is write-only..."
        logger.error(msg)
        raise NotImplementedError(msg)

    @recall.setter
    def recall(self, value: int) -> None:
        """Recall the panel settings from a given slot. Available slot: 1...5

        :param value: `int` between 1 and 5 to select the slot
        """
        validate_number("recall", value, (1, 5), number_type=int, logger=logger)
        value_str = f"RCL{value}"
        logger.debug(f'Recall a setting from slot {value} with: "{value_str}"')
        self.com.write(value_str)

    @contextlib.contextmanager
    def save_settings_to(self, slot: int) -> Generator[None, Any, None]:
        """Method to save settings to a slot.

        Use this method with a `with`-block. It will recall a slot and save the settings
        to it.

        Example:

        .. code-block:: python

            slot = 3
            with ka3005p.save_settings_to(slot):
                ka3005p.voltage = 2.5
                ka3005p.current = 0.75

        :param slot: Slot 1...5

        """
        self.recall = slot
        yield
        sleep(self.config.time_to_wait_before_save_settings)
        self.save = slot

    # OCP
    @property
    def ocp(self) -> bool:
        """State of the over current protection"""
        return self.status.ocp

    @ocp.setter
    def ocp(self, value: bool) -> None:
        """
        Activates the over current protection

        :param value: `True` for activation, `False` for deactivation
        """
        validate_bool("ocp", value, logger=logger)
        value_str = f"OCP{1 if value else 0}"
        logger.debug(
            f"Set over current protection ({'ON' if value else 'OFF'}) "
            f'with: "{value_str}"'
        )
        self.com.write(value_str)

    # OVP
    @property
    def ovp(self) -> bool:
        """State of the over voltage protection"""
        return self.status.ovp

    @ovp.setter
    def ovp(self, value: bool) -> None:
        """
        Activates the over voltage protection

        :param value: `True` for activation, `False` for deactivation
        """
        validate_bool("ovp", value, logger=logger)
        value_str = f"OVP{1 if value else 0}"
        logger.debug(
            f"Set over voltage protection ({'ON' if value else 'OFF'}) "
            f'with: "{value_str}"'
        )
        self.com.write(value_str)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Mock Oscilloscope
"""

from array import array

from libtiepie import const
from libtiepie.oscilloscope import Oscilloscope as LtpOscilloscope

from .const import MOCK_OSCILLOSCOPE_SERIAL_NUMBER_2
from .oscilloscopechannel import OscilloscopeChannel
from .oscilloscopetrigger import OscilloscopeTrigger


class Oscilloscope(LtpOscilloscope):
    """"""

    def __init__(self, serial_number) -> None:
        self._serial_number = serial_number

        self._is_running = False
        self._sample_rate = None
        self._record_length = None
        self._pre_sample_ratio = None
        self._valid_pre_sample_count = None
        self._resolution = None
        self._trigger = OscilloscopeTrigger()
        self._is_data_ready = True
        self._mock_reduce_data = 0
        self._is_triggered = False
        self._auto_resolution_mode = const.ARM_UNKNOWN

        self._channels = []

        # add 3 test channels
        self.add_channel()
        self.add_channel()
        self.add_channel()

        # add physical device parameters
        self._record_length_max = 1e9
        self._sample_rate_max = 1e9
        self._trigger_delay_max = 1
        if serial_number == MOCK_OSCILLOSCOPE_SERIAL_NUMBER_2:
            self._block_measurement_support = False
        else:
            self._block_measurement_support = True

    def add_channel(self) -> None:
        channel = OscilloscopeChannel()
        self._channels.append(channel)

    def _get_channels(self):
        return self._channels

    def _get_trigger(self):
        return self._trigger

    def _get_block_measurement_support(self):
        return self._block_measurement_support

    def _get_measure_modes(self):
        return const.MM_BLOCK if self._block_measurement_support else 0

    def _get_pre_sample_ratio(self):
        return self._pre_sample_ratio

    def _set_pre_sample_ratio(self, value):
        self._pre_sample_ratio = value

    def _get_record_length(self):
        return self._record_length

    def _get_record_length_max(self):
        return self._record_length_max

    def _set_record_length(self, value):
        self._record_length = value

    def _get_valid_pre_sample_count(self):
        return self._valid_pre_sample_count

    def _set_valid_presample_count(self, value):
        self._valid_pre_sample_count = value

    def verify_record_length(self, record_length):
        return record_length

    def _get_resolution(self):
        return self._resolution

    def _set_resolution(self, value):
        self._resolution = value

    def _get_auto_resolution_mode(self):
        return self._auto_resolution_mode

    def _set_auto_resolution_mode(self, value):
        self._auto_resolution_mode = value

    def _get_trigger_delay_max(self):
        return self._trigger_delay_max

    def _get_sample_rate(self):
        return self._sample_rate

    def _set_sample_rate(self, value):
        self._sample_rate = value

    def verify_sample_rate(self, sample_rate):
        return sample_rate

    def _get_sample_rate_max(self):
        return self._sample_rate_max

    def verify_trigger_time_out(self, timeout):
        return timeout

    def _get_is_data_ready(self):
        return self._is_data_ready

    def _set_is_data_ready(self, is_data_ready):
        self._is_data_ready = is_data_ready

    def _get_mock_reduce_data(self):
        return self._mock_reduce_data

    def _set_mock_reduce_data(self, mock_reduce_data):
        self._mock_reduce_data = mock_reduce_data

    def _get_is_triggered(self):
        return self._is_triggered

    def _set_is_triggered(self, is_triggered):
        self._is_triggered = is_triggered

    def _get_is_running(self):
        return self._is_running

    def _set_is_running(self, is_running):
        self._is_running = is_running

    def force_trigger(self) -> None:
        pass

    def start(self) -> bool:
        return True

    def stop(self) -> None:
        pass

    def get_data(self):
        data = [
            (
                array("d", range(self.record_length - 1 - self.mock_reduce_data))
                if ch.enabled
                else None
            )
            # Generate measured data; length is reduced by 1 to mock a non-completely
            # filled buffer
            for _, ch in enumerate(self.channels)
        ]
        self.mock_reduce_data = 0
        return data

    def __del__(self) -> None:
        pass

    measure_modes = property(_get_measure_modes)
    pre_sample_ratio = property(_get_pre_sample_ratio, _set_pre_sample_ratio)
    record_length = property(_get_record_length, _set_record_length)
    valid_pre_sample_count = property(
        _get_valid_pre_sample_count, _set_valid_presample_count
    )
    resolution = property(_get_resolution, _set_resolution)
    auto_resolution_mode = property(
        _get_auto_resolution_mode,
        _set_auto_resolution_mode,
    )
    sample_rate = property(_get_sample_rate, _set_sample_rate)
    channels = property(_get_channels)
    record_length_max = property(_get_record_length_max)
    sample_rate_max = property(_get_sample_rate_max)
    trigger_delay_max = property(_get_trigger_delay_max)

    is_running = property(_get_is_running, _set_is_running)
    is_data_ready = property(_get_is_data_ready, _set_is_data_ready)
    is_triggered = property(_get_is_triggered, _set_is_triggered)
    mock_reduce_data = property(_get_mock_reduce_data, _set_mock_reduce_data)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Tests for the .dev.crylas sub-package.
"""

import logging
import time

import pytest

from hvl_ccb.comm.serial import SerialCommunicationIOError
from hvl_ccb.dev.crylas import (
    CryLasAttenuator,
    CryLasAttenuatorError,
    CryLasLaser,
    CryLasLaserConfig,
    CryLasLaserError,
    CryLasLaserNotReadyError,
    CryLasLaserSerialCommunication,
    CryLasLaserSerialCommunicationConfig,
)
from masked_comm.serial import (
    CryLasLaserLoopSerialCommunication,
    LoopSerialCommunication,
)

logging.basicConfig(level=logging.ERROR)


@pytest.fixture(scope="module")
def laser_com_config():
    return {
        "port": "loop://?logging=debug",
        "baudrate": 19200,
        "parity": CryLasLaserSerialCommunicationConfig.Parity.NONE,
        "stopbits": CryLasLaserSerialCommunicationConfig.Stopbits.ONE,
        "bytesize": CryLasLaserSerialCommunicationConfig.Bytesize.EIGHTBITS,
        "terminator": b"\n",
        "timeout": 0.1,
    }


@pytest.fixture(scope="module")
def laser_config():
    return {
        "polling_period": 0.05,
        "auto_laser_on": False,
        "init_shutter_status": CryLasLaserConfig.ShutterStatus.CLOSED,
    }


@pytest.fixture
def started_laser(laser_com_config, laser_config):
    com = CryLasLaserLoopSerialCommunication(laser_com_config)
    com.open()
    com.put_text(CryLasLaser.AnswersShutter.CLOSED.value)
    com.put_text(CryLasLaser.AnswersStatus.READY.value)
    com.put_text("")
    with CryLasLaser(com, laser_config) as las:
        while com.get_written() is not None:
            pass
        yield com, las


def test_laser_com(laser_com_config) -> None:
    com = CryLasLaserLoopSerialCommunication(laser_com_config)
    # test error when com closed
    assert not com.is_open
    with pytest.raises(SerialCommunicationIOError):
        com.query("attempting to do a query while com is closed", "prefix")
    # the com still gets written
    com.get_written()
    with pytest.raises(SerialCommunicationIOError):
        com.query_all("attempting to do a query while com is closed", "prefix")
    # the com still gets written
    com.get_written()
    # test query and query_all work when com opened
    com.open()
    com.put_text("ignore this first")
    com.put_text("prefix hello")
    assert com.query("query", "prefix", "post query command") == "prefix hello"
    assert com.get_written() == "query"
    assert com.get_written() == "post query command"
    com.put_text("ignore this first")
    com.put_text("prefix hello")
    com.put_text("prefix hi")
    com.put_text("prefix welcome")
    com.put_text(f"{CryLasLaserSerialCommunication.READ_TEXT_SKIP_PREFIXES[0]} nah")
    com.put_text("")
    assert com.query_all("query", "prefix") == [
        "prefix hello",
        "prefix hi",
        "prefix welcome",
    ]
    assert com.get_written() == "query"
    com.close()


def test_laser_instantiation(laser_com_config, laser_config) -> None:
    las = CryLasLaser(laser_com_config)
    assert las is not None

    wrong_config = dict(laser_config)
    wrong_config["polling_period"] = [-1]
    with pytest.raises(TypeError):
        CryLasLaser(laser_com_config, wrong_config)
    wrong_config["polling_period"] = -1
    with pytest.raises(ValueError):
        CryLasLaser(laser_com_config, wrong_config)

    wrong_config = dict(laser_config)
    wrong_config["calibration_factor"] = None
    with pytest.raises(TypeError):
        CryLasLaser(laser_com_config, wrong_config)
    wrong_config["calibration_factor"] = -0.1
    with pytest.raises(ValueError):
        CryLasLaser(laser_com_config, wrong_config)
    wrong_config["calibration_factor"] = 0
    with pytest.raises(ValueError):
        CryLasLaser(laser_com_config, wrong_config)

    wrong_config = dict(laser_config)
    wrong_config["polling_timeout"] = "NaN"
    with pytest.raises(TypeError):
        CryLasLaser(laser_com_config, wrong_config)
    wrong_config["polling_timeout"] = -0.5
    with pytest.raises(ValueError):
        CryLasLaser(laser_com_config, wrong_config)
    wrong_config["polling_timeout"] = 0
    with pytest.raises(ValueError):
        CryLasLaser(laser_com_config, wrong_config)

    wrong_config = dict(laser_config)
    wrong_config["init_shutter_status"] = ()
    with pytest.raises(TypeError):
        CryLasLaser(laser_com_config, wrong_config)
    wrong_config["init_shutter_status"] = 2
    with pytest.raises(ValueError):
        CryLasLaser(laser_com_config, wrong_config)


def test_laser_start(laser_com_config, laser_config) -> None:
    com = CryLasLaserLoopSerialCommunication(laser_com_config)
    las = CryLasLaser(com, laser_config)
    com.open()
    # starting again should work
    com.put_text(CryLasLaser.AnswersShutter.CLOSED.value)
    com.put_text(las.AnswersStatus.READY.value)
    com.put_text("")
    las.start()
    assert las.laser_status == las.LaserStatus.READY_INACTIVE
    # test case: system not immediately ready, execution not blocked
    las.laser_status = las.LaserStatus.UNREADY_INACTIVE
    com.put_text(CryLasLaser.AnswersShutter.CLOSED.value)
    com.put_text(las.AnswersStatus.READY.value)
    com.put_text("")
    las.start()
    com.put_text(CryLasLaser.AnswersShutter.CLOSED.value)
    las.close_shutter()
    # waiting for the polling thread
    las.wait_until_ready()
    assert las.laser_status == las.LaserStatus.READY_INACTIVE
    # teardown
    com.put_text(las.AnswersStatus.INACTIVE.value)
    com.put_text(las.AnswersShutter.CLOSED.value)
    las.stop()

    # error on input jibrish
    com = CryLasLaserLoopSerialCommunication(laser_com_config)
    las = CryLasLaser(com, laser_config)
    com.open()
    com.put_text(CryLasLaser.AnswersShutter.CLOSED.value)
    com.put_text("STATUS: Not one of expected answers")
    with pytest.raises(ValueError):
        las.start()


def test_laser_wait_until_ready(laser_com_config, laser_config) -> None:
    # test case: system not immediately ready, execution blocked
    com = CryLasLaserLoopSerialCommunication(laser_com_config)

    las = CryLasLaser(com, laser_config)
    com.open()
    com.put_text(CryLasLaser.AnswersShutter.CLOSED.value)
    com.put_text(las.AnswersStatus.HEAD.value)
    com.put_text("")
    assert las.laser_status == las.LaserStatus.UNREADY_INACTIVE
    las.start()
    assert las.laser_status == las.LaserStatus.UNREADY_INACTIVE
    com.put_text(las.AnswersStatus.READY.value)
    com.put_text("")
    las.wait_until_ready()
    assert las.laser_status == las.LaserStatus.READY_INACTIVE
    com.put_text(CryLasLaser.AnswersShutter.CLOSED.value)
    las.close_shutter()
    com.put_text(CryLasLaser.AnswersStatus.INACTIVE.value)
    com.put_text(CryLasLaser.AnswersShutter.CLOSED.value)
    las.stop()
    assert not com.is_open

    # wait on start
    laser_config_dict = dict(laser_config)
    las = CryLasLaser(com, laser_config_dict)
    com.open()
    com.put_text(las.AnswersShutter.CLOSED.value)
    com.put_text(las.AnswersStatus.READY.value)
    com.put_text("")
    assert las.laser_status == las.LaserStatus.UNREADY_INACTIVE
    las.start()
    assert las.laser_status != las.LaserStatus.UNREADY_INACTIVE
    com.put_text(las.AnswersShutter.CLOSED.value)
    las.close_shutter()
    com.put_text(las.AnswersStatus.INACTIVE.value)
    com.put_text(las.AnswersShutter.CLOSED.value)
    las.stop()

    # no wait if already ready
    las = CryLasLaser(com, laser_config)
    com.open()
    com.put_text(las.AnswersShutter.CLOSED.value)
    com.put_text(las.AnswersStatus.READY.value)
    las.start()
    assert las.laser_status != las.LaserStatus.UNREADY_INACTIVE
    las.wait_until_ready()
    assert las.laser_status != las.LaserStatus.UNREADY_INACTIVE
    com.put_text(las.AnswersStatus.INACTIVE.value)
    com.put_text(las.AnswersShutter.CLOSED.value)
    las.stop()


def test_status_polling_end(laser_com_config, laser_config) -> None:
    # test case: system not immediately ready, execution blocked
    com = CryLasLaserLoopSerialCommunication(laser_com_config)

    # stop status polling on device stop
    laser_config_dict = dict(laser_config)
    # set polling period big enough for stop() to trigger before inactive status is read
    laser_config_dict["polling_period"] = 0.2
    las = CryLasLaser(com, laser_config_dict)
    assert las.laser_status == las.LaserStatus.UNREADY_INACTIVE
    com.open()
    com.put_text(las.AnswersShutter.CLOSED.value)
    com.put_text(las.AnswersStatus.HEAD.value)
    com.put_text("")
    las.start()
    las._start_polling()  # BLAH: call internal (other are blocking)
    time.sleep(0.1)  # let the polling thread startup
    assert las._is_polling()  # BLAH: call internal (make public?)
    com.put_text(las.AnswersStatus.INACTIVE.value)
    com.put_text(las.AnswersShutter.CLOSED.value)
    las.stop()
    assert las.laser_status == las.LaserStatus.READY_INACTIVE

    # status polling timeout error
    laser_config_dict = dict(laser_config)
    # set polling period big enough
    laser_config_dict["polling_period"] = 0.2
    laser_config_dict["polling_timeout"] = laser_config_dict["polling_period"] / 10
    las = CryLasLaser(com, laser_config_dict)
    assert las.laser_status == las.LaserStatus.UNREADY_INACTIVE
    com.open()
    com.put_text(las.AnswersShutter.CLOSED.value)
    com.put_text(las.AnswersStatus.HEAD.value)
    las.start()
    assert las.laser_status == las.LaserStatus.UNREADY_INACTIVE
    with pytest.raises(CryLasLaserError):
        las.wait_until_ready()
    com.put_text(las.AnswersStatus.INACTIVE.value)
    com.put_text(las.AnswersShutter.CLOSED.value)
    las.stop()
    assert las.laser_status == las.LaserStatus.READY_INACTIVE


def test_laser_stop(started_laser) -> None:
    com, las = started_laser
    com.put_text(las.AnswersStatus.INACTIVE.value)
    com.put_text(las.AnswersShutter.CLOSED.value)
    las.stop()
    assert not com.is_open
    assert las.laser_status == las.LaserStatus.READY_INACTIVE
    assert las.shutter_status == las.ShutterStatus.CLOSED
    # stop is an idempotent operation
    las.stop()
    assert not com.is_open
    assert las.laser_status == las.LaserStatus.READY_INACTIVE
    assert las.shutter_status == las.ShutterStatus.CLOSED


def test_laser_on(started_laser, laser_config) -> None:
    com, las = started_laser
    # normal case
    com.put_text(las.AnswersStatus.READY.value)
    com.put_text(las.AnswersStatus.ACTIVE.value)
    com.put_text("")
    assert las.laser_status == las.LaserStatus.READY_INACTIVE
    las.laser_on()
    assert com.get_written() == "LASER ON"
    assert las.laser_status == las.LaserStatus.READY_ACTIVE
    # case laser not ready
    las.laser_status = las.LaserStatus.UNREADY_INACTIVE
    with pytest.raises(CryLasLaserNotReadyError):
        las.laser_on()
    # case laser already on
    las.laser_status = las.LaserStatus.READY_ACTIVE
    com.put_text("Laser is already on")
    las.laser_on()
    # case command fails
    las.laser_status = las.LaserStatus.READY_INACTIVE
    com.put_text("The command is not working")
    with pytest.raises(CryLasLaserError):
        las.laser_on()
    # teardown
    com.put_text(las.AnswersStatus.INACTIVE.value)
    com.put_text(las.AnswersShutter.CLOSED.value)
    las.stop()
    assert not com.is_open

    laser_config_dict = dict(laser_config)
    laser_config_dict["auto_laser_on"] = True
    las = CryLasLaser(com, laser_config_dict)
    assert las.laser_status == las.LaserStatus.UNREADY_INACTIVE
    com.open()
    com.put_text(las.AnswersShutter.CLOSED.value)
    com.put_text(las.AnswersStatus.READY.value)
    com.put_text("")
    com.put_text(las.AnswersStatus.ACTIVE.value)
    las.start()
    las.wait_until_ready()
    assert las.laser_status == las.LaserStatus.READY_ACTIVE
    com.put_text(las.AnswersStatus.ACTIVE.value)
    las.laser_on()
    assert las.laser_status == las.LaserStatus.READY_ACTIVE
    com.put_text(las.AnswersStatus.INACTIVE.value)
    com.put_text(las.AnswersShutter.CLOSED.value)
    las.stop()
    assert not com.is_open

    laser_config_dict = dict(laser_config)
    laser_config_dict["auto_laser_on"] = True
    las = CryLasLaser(com, laser_config_dict)
    assert las.laser_status == las.LaserStatus.UNREADY_INACTIVE
    com.open()
    com.put_text(las.AnswersShutter.CLOSED.value)
    com.put_text(las.AnswersStatus.HEAD.value)
    com.put_text("")
    com.put_text(las.AnswersStatus.READY.value)
    com.put_text("")
    com.put_text(las.AnswersStatus.ACTIVE.value)
    las.start()
    las.wait_until_ready()
    assert las.laser_status == las.LaserStatus.READY_ACTIVE
    com.put_text(las.AnswersStatus.ACTIVE.value)
    com.put_text("")
    las.laser_on()
    assert las.laser_status == las.LaserStatus.READY_ACTIVE
    com.put_text(las.AnswersStatus.INACTIVE.value)
    com.put_text(las.AnswersShutter.CLOSED.value)


def test_laser_off(started_laser) -> None:
    com, las = started_laser
    # normal case
    las.laser_status = las.LaserStatus.READY_ACTIVE
    com.put_text(las.AnswersStatus.INACTIVE.value)
    las.laser_off()
    assert com.get_written() == "LASER OFF"
    assert las.laser_status == las.LaserStatus.READY_INACTIVE
    # case laser already off
    las.laser_status = las.LaserStatus.READY_INACTIVE
    com.put_text("Laser is already off")
    las.laser_off()
    # case command fails
    las.laser_status = las.LaserStatus.READY_ACTIVE
    com.put_text("The command is not working")
    with pytest.raises(CryLasLaserError):
        las.laser_off()
    # teardown
    com.put_text(las.AnswersStatus.INACTIVE.value)
    com.put_text(las.AnswersShutter.CLOSED.value)


def test_laser_open_shutter(started_laser) -> None:
    com, las = started_laser
    # normal case
    com.put_text(las.AnswersShutter.OPENED.value)
    las.open_shutter()
    assert com.get_written() == "Shutter 1"
    assert las.shutter_status == las.ShutterStatus.OPENED
    # case command fails
    com.put_text("The command is not working")
    with pytest.raises(CryLasLaserError):
        las.open_shutter()
    # teardown
    com.put_text(las.AnswersStatus.INACTIVE.value)
    com.put_text(las.AnswersShutter.CLOSED.value)


def test_laser_close_shutter(started_laser) -> None:
    com, las = started_laser
    # normal case
    com.put_text(las.AnswersShutter.CLOSED.value)
    las.close_shutter()
    assert com.get_written() == "Shutter 0"
    assert las.shutter_status == las.ShutterStatus.CLOSED
    # case command fails
    com.put_text("The command is not working")
    with pytest.raises(CryLasLaserError):
        las.close_shutter()
    # teardown
    com.put_text(las.AnswersStatus.INACTIVE.value)
    com.put_text(las.AnswersShutter.CLOSED.value)


def test_laser_set_init_shutter_status(laser_com_config, laser_config) -> None:
    # test with init CLOSED
    com = CryLasLaserLoopSerialCommunication(laser_com_config)
    las = CryLasLaser(com, laser_config)
    # test when command fails
    com.open()
    com.put_text("")
    with pytest.raises(CryLasLaserError):
        las.set_init_shutter_status()
    # test when command succeeds and there is nothing to do
    com.put_text(las.AnswersShutter.CLOSED.value)
    las.set_init_shutter_status()
    # test when command succeeds and the laser should be closed
    com.put_text(las.AnswersShutter.OPENED.value)
    com.put_text(las.AnswersShutter.CLOSED.value)
    las.set_init_shutter_status()
    # teardown
    com.put_text(las.AnswersStatus.INACTIVE.value)
    com.put_text(las.AnswersShutter.CLOSED.value)
    las.stop()
    # test with init OPENED
    com = CryLasLaserLoopSerialCommunication(laser_com_config)
    laser_config["init_shutter_status"] = CryLasLaserConfig.ShutterStatus.OPENED
    las = CryLasLaser(com, laser_config)
    com.open()
    # test when command succeeds and there is nothing to do
    com.put_text(las.AnswersShutter.OPENED.value)
    las.set_init_shutter_status()
    # test when command succeeds and the laser should be opened
    com.put_text(las.AnswersShutter.CLOSED.value)
    com.put_text(las.AnswersShutter.OPENED.value)
    las.set_init_shutter_status()
    # teardown
    laser_config["init_shutter_status"] = CryLasLaserConfig.ShutterStatus.CLOSED
    com.put_text(las.AnswersStatus.INACTIVE.value)
    com.put_text(las.AnswersShutter.CLOSED.value)


def test_laser_update_laser_status(started_laser) -> None:
    com, las = started_laser
    assert las.laser_status == las.LaserStatus.READY_INACTIVE
    # no answer error
    com.put_text("")
    with pytest.raises(CryLasLaserError):
        las.update_laser_status()
    # unexpected answer error
    com.put_text(las.AnswersStatus.TEC1.value)
    com.put_text(las.AnswersStatus.HEAD.value)
    com.put_text(las.AnswersStatus.INACTIVE.value)
    com.put_text(las.AnswersStatus.READY.value)
    com.put_text("")
    with pytest.raises(CryLasLaserError):
        las.update_laser_status()
    # ignore some lines and updated status
    com.put_text(las.AnswersShutter.OPENED.value)
    com.put_text(las.AnswersStatus.HEAD.value)
    com.put_text("")
    las.update_laser_status()
    assert las.laser_status == las.LaserStatus.UNREADY_INACTIVE
    com.put_text(las.AnswersStatus.READY.value)
    com.put_text("")
    las.update_laser_status()
    assert las.laser_status == las.LaserStatus.READY_INACTIVE
    com.put_text(las.AnswersStatus.ACTIVE.value)
    com.put_text("")
    las.update_laser_status()
    assert las.laser_status == las.LaserStatus.READY_ACTIVE
    com.put_text(las.AnswersStatus.INACTIVE.value)
    com.put_text("")
    with pytest.raises(CryLasLaserError):
        las.update_laser_status()
    las.laser_status = las.LaserStatus.UNREADY_INACTIVE
    com.put_text(las.AnswersStatus.HEAD.value)
    com.put_text("")
    las.update_laser_status()
    assert las.laser_status == las.LaserStatus.UNREADY_INACTIVE
    # teardown
    com.put_text(las.AnswersStatus.INACTIVE.value)
    com.put_text(las.AnswersShutter.CLOSED.value)


def test_laser_update_shutter_status(started_laser) -> None:
    com, las = started_laser
    com.put_text(las.AnswersShutter.CLOSED.value)
    las.update_shutter_status()
    assert las.shutter_status == las.ShutterStatus.CLOSED
    com.put_text(las.AnswersShutter.OPENED.value)
    las.update_shutter_status()
    assert las.shutter_status == las.ShutterStatus.OPENED
    com.put_text("")
    with pytest.raises(CryLasLaserError):
        las.update_shutter_status()
    # teardown
    com.put_text(las.AnswersStatus.INACTIVE.value)
    com.put_text(las.AnswersShutter.CLOSED.value)


def test_laser_update_repetition_rate(started_laser) -> None:
    com, las = started_laser
    com.put_text("Impuls=enabled 10Hz")
    las.update_repetition_rate()
    assert las.repetition_rate == las.RepetitionRates.SOFTWARE_INTERNAL_TEN
    com.put_text("Impuls=disabled, extern Trigger")
    las.update_repetition_rate()
    assert las.repetition_rate == las.RepetitionRates.HARDWARE
    com.put_text("no answer to query")
    with pytest.raises(CryLasLaserError):
        las.update_repetition_rate()
    com.put_text("Impuls=enabled but no number provided")
    with pytest.raises(CryLasLaserError):
        las.update_repetition_rate()
    # teardown
    com.put_text(las.AnswersStatus.INACTIVE.value)
    com.put_text(las.AnswersShutter.CLOSED.value)


def test_laser_update_target_pulse_energy(started_laser) -> None:
    com, las = started_laser
    com.put_text("PD-Sollwert=46000")
    las.update_target_pulse_energy()
    assert las.target_pulse_energy == int(46000 * las.config.calibration_factor / 1000)
    com.put_text("")
    with pytest.raises(CryLasLaserError):
        las.update_target_pulse_energy()
    # teardown
    com.put_text(las.AnswersStatus.INACTIVE.value)
    com.put_text(las.AnswersShutter.CLOSED.value)


def test_laser_get_pulse_energy_and_rate(started_laser) -> None:
    com, las = started_laser
    # test that the command works
    com.put_text(
        "IST: LDTemp=0, NLOTemp=1, CaseTemp=2,"
        " PD=46000, D_I= 4, freq=0, LDE=5, RegOFF=7"
    )
    assert las.get_pulse_energy_and_rate() == (200, 0)
    assert com.get_written() == "DB1"
    assert com.get_written() == "DB0"
    # test that the command fails
    com.put_text("")
    with pytest.raises(CryLasLaserError):
        las.get_pulse_energy_and_rate()
    com.put_text(las.AnswersStatus.INACTIVE.value)
    com.put_text(las.AnswersShutter.CLOSED.value)


def test_laser_set_repetition_rate(started_laser) -> None:
    com, las = started_laser
    # check param type int
    com.put_text("Impuls=enabled 10Hz")
    las.set_repetition_rate(10)
    assert las.repetition_rate.value == 10
    assert com.get_written() == f"BOO IP{las.RepetitionRates(10).send_value}"
    # check param type RepetitionRates
    com.put_text("Impuls=enabled 20Hz")
    las.set_repetition_rate(las.RepetitionRates(20))
    assert las.repetition_rate == las.RepetitionRates(20)
    assert com.get_written() == f"BOO IP{las.RepetitionRates(20).send_value}"
    # check value Error
    with pytest.raises(ValueError):
        las.set_repetition_rate(30)
    # command fails
    com.put_text("This did not work")
    with pytest.raises(CryLasLaserError):
        las.set_repetition_rate(10)
    # hardware trigger
    com.put_text("No answer is expected in this case")
    las.set_repetition_rate(0)
    # teardown
    com.put_text(las.AnswersStatus.INACTIVE.value)
    com.put_text(las.AnswersShutter.CLOSED.value)


def test_laser_set_pulse_energy(started_laser) -> None:
    com, las = started_laser
    # case successful
    val = 100
    cmd_val = int(val * 1000 / las.config.calibration_factor)
    com.put_text(f"PD-Sollwert={cmd_val}")
    las.set_pulse_energy(val)
    assert las.target_pulse_energy == int(
        cmd_val * las.config.calibration_factor / 1000
    )
    assert com.get_written() == f"BOO SE {cmd_val}"
    # case unsuccessful
    val = 100
    cmd_val = int(val * 1000 / las.config.calibration_factor)
    com.put_text("PD-Sollwert=46000")
    with pytest.raises(CryLasLaserError):
        las.set_pulse_energy(val)
    assert com.get_written() == f"BOO SE {cmd_val}"
    # case command fails
    com.put_text("The command failed")
    with pytest.raises(CryLasLaserError):
        las.set_pulse_energy(val)
    # teardown
    com.put_text(las.AnswersStatus.INACTIVE.value)
    com.put_text(las.AnswersShutter.CLOSED.value)


def test_laser_com_error(laser_com_config, laser_config) -> None:
    wrong_config = dict(laser_com_config)
    wrong_config["port"] = "NOT A PORT"
    las = CryLasLaser(wrong_config, laser_config)
    assert not las.com.is_open

    with pytest.raises(SerialCommunicationIOError):
        las.start()

    las = CryLasLaser(laser_com_config, laser_config)
    assert not las.com.is_open

    assert las.laser_status is las.LaserStatus.UNREADY_INACTIVE
    with pytest.raises(SerialCommunicationIOError):
        las.wait_until_ready()
    las.laser_status = las.LaserStatus.READY_INACTIVE
    with pytest.raises(SerialCommunicationIOError):
        las.laser_on()
    with pytest.raises(SerialCommunicationIOError):
        las.laser_off()
    with pytest.raises(SerialCommunicationIOError):
        las.open_shutter()
    with pytest.raises(SerialCommunicationIOError):
        las.close_shutter()
    with pytest.raises(SerialCommunicationIOError):
        las.set_init_shutter_status()
    with pytest.raises(SerialCommunicationIOError):
        las.update_laser_status()
    with pytest.raises(SerialCommunicationIOError):
        las.update_shutter_status()
    with pytest.raises(SerialCommunicationIOError):
        las.update_repetition_rate()
    with pytest.raises(SerialCommunicationIOError):
        las.update_target_pulse_energy()
    with pytest.raises(SerialCommunicationIOError):
        las.get_pulse_energy_and_rate()
    with pytest.raises(SerialCommunicationIOError):
        las.set_repetition_rate(CryLasLaser.RepetitionRates.SOFTWARE_INTERNAL_TEN)
    with pytest.raises(SerialCommunicationIOError):
        las.set_pulse_energy(200)


def test_auto_laser_on(laser_com_config, laser_config) -> None:
    com = CryLasLaserLoopSerialCommunication(laser_com_config)

    config_where_auto_laser_is_on = dict(laser_config)
    config_where_auto_laser_is_on["auto_laser_on"] = True

    las = CryLasLaser(com, config_where_auto_laser_is_on)

    com.open()

    # answers to queries made during start()
    com.put_text(CryLasLaser.AnswersShutter.CLOSED.value)
    com.put_text(las.AnswersStatus.HEAD.value)
    com.put_text("")

    # check that the laser status is inactive before start()
    assert las.laser_status == las.LaserStatus.UNREADY_INACTIVE

    las.start()

    # queries that should have been made during start
    assert com.get_written() == "Shutter"
    assert com.get_written() == "STATUS"

    com.put_text(las.AnswersStatus.READY.value)
    com.put_text(las.AnswersStatus.ACTIVE.value)
    com.put_text("")

    time.sleep(1)
    assert com.get_written() == "STATUS"
    assert com.get_written() == "LASER ON"

    assert las.laser_status == las.LaserStatus.READY_ACTIVE


# tests attenuator


@pytest.fixture(scope="module")
def attenuator_com_config():
    return {
        "port": "loop://?logging=debug",
        "baudrate": 9600,
        "parity": CryLasLaserSerialCommunicationConfig.Parity.NONE,
        "stopbits": CryLasLaserSerialCommunicationConfig.Stopbits.ONE,
        "bytesize": CryLasLaserSerialCommunicationConfig.Bytesize.EIGHTBITS,
        "terminator": b"",
        "timeout": 0.1,
    }


@pytest.fixture(scope="module")
def attenuator_config():
    return {
        "init_attenuation": 0,
        "response_sleep_time": 0.01,
    }


@pytest.fixture
def started_attenuator(attenuator_com_config, attenuator_config):
    serial_port = LoopSerialCommunication(attenuator_com_config)
    serial_port.open()
    serial_port.put_bytes(bytes([0, 0]))
    with CryLasAttenuator(serial_port, attenuator_config) as att:
        while serial_port.get_written() is not None:
            pass
        yield serial_port, att


def test_attenuator_instantiation(attenuator_com_config, attenuator_config) -> None:
    att = CryLasAttenuator(attenuator_com_config)
    assert att is not None

    wrong_config = dict(attenuator_config)
    wrong_config["init_attenuation"] = -1
    with pytest.raises(ValueError):
        CryLasAttenuator(attenuator_com_config, wrong_config)

    wrong_config = dict(attenuator_config)
    wrong_config["response_sleep_time"] = 0
    with pytest.raises(ValueError):
        CryLasAttenuator(attenuator_com_config, wrong_config)
    wrong_config["response_sleep_time"] = -0.1
    with pytest.raises(ValueError):
        CryLasAttenuator(attenuator_com_config, wrong_config)


def test_attenuator_start(started_attenuator) -> None:
    com, att = started_attenuator
    com.put_bytes(bytes([0, 0]))
    att.start()
    assert com.get_written()


def test_attenuator_set_attenuation(started_attenuator) -> None:
    com, att = started_attenuator
    com.put_bytes(bytes([0, 0]))
    att.set_attenuation(50)
    assert att.attenuation == 50
    com.put_bytes(bytes([1, 0]))
    with pytest.raises(CryLasAttenuatorError):
        att.set_attenuation(50)
    with pytest.raises(ValueError):
        att.set_attenuation(-1)


def test_attenuator_com_error(attenuator_com_config, attenuator_config) -> None:
    wrong_config = dict(attenuator_com_config)
    wrong_config["port"] = "NOT A PORT"
    att = CryLasAttenuator(wrong_config, attenuator_config)
    assert not att.com.is_open
    with pytest.raises(SerialCommunicationIOError):
        att.start()

    att = CryLasAttenuator(attenuator_com_config, attenuator_config)
    assert not att.com.is_open
    with pytest.raises(SerialCommunicationIOError):
        att.set_attenuation(50)


def test_attenuator_set_init_attenuation(started_attenuator) -> None:
    com, att = started_attenuator
    com.put_bytes(bytes([0, 0]))
    att.set_init_attenuation()
    assert att.attenuation == att.config.init_attenuation


def test_attenuator_set_transmission(started_attenuator) -> None:
    com, att = started_attenuator
    com.put_bytes(bytes([0, 0]))
    att.set_transmission(50)
    assert att.transmission == 50
    com.put_bytes(bytes([1, 0]))
    with pytest.raises(CryLasAttenuatorError):
        att.set_transmission(50)
    with pytest.raises(ValueError):
        att.set_transmission(-1)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#

import logging
from typing import NamedTuple

from bitstring import BitArray

from hvl_ccb.dev.base import DeviceError
from hvl_ccb.utils.enum import NameEnum
from hvl_ccb.utils.validation import validate_number

logger = logging.getLogger(__name__)


class KA3000PError(DeviceError):
    """KA3000P related errors."""


class KA3000PTracking(NameEnum):
    SINGLE = 0
    INDEPENDENT = 1
    PARALLEL = 2
    SERIAL = 3


class KA3000PStatus(NamedTuple):
    ch1_cv: bool
    ch2_cv: bool
    tracking: KA3000PTracking
    beep: bool
    ocp: bool
    output: bool
    ovp: bool


def _gen_status(stb_qry: int) -> KA3000PStatus:
    """Converts a sinlge byte into the `KA3000PStatus`

    :param stb_qry: status as single byte
    :return: Status of KA3000P
    """
    validate_number("stb_qry", stb_qry, (0, 255), int, logger=logger)
    stb = BitArray(uint=stb_qry, length=8)
    stb.reverse()
    ch1_cv = bool(stb[0])
    ch2_cv = bool(stb[1])
    tracking = KA3000PTracking(stb[2:4].uint)
    beep = bool(stb[4])
    ocp = bool(stb[5])
    output = bool(stb[6])
    ovp = bool(stb[7])

    return KA3000PStatus(ch1_cv, ch2_cv, tracking, beep, ocp, output, ovp)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Test module for the BaseCube device class.

NOTE: After a variable is set for the first time, the type is fixed! This can usually
happen for numerical values. Therefore, we should always use 'float' as datatype for
numerical variables.

GOOD:
demo_opcua_server.set_var(constants.Power.voltage_target, 0.0)

BAD:
demo_opcua_server.set_var(constants.Power.voltage_target, 0)
"""

import logging
import sys
from concurrent.futures import ThreadPoolExecutor
from itertools import product
from time import sleep

import pytest
from pytest_mock import MockerFixture

from hvl_ccb.comm.opc import OpcUaCommunicationIOError, OpcUaCommunicationTimeoutError
from hvl_ccb.dev.cube import (
    BaseCube,
    BaseCubeConfiguration,
    BaseCubeOpcUaCommunication,
    CubeEarthingStickOperationError,
    CubeRemoteControlError,
    CubeStatusChangeError,
    CubeStopError,
    alarms,
    constants,
    earthing_stick,
)
from hvl_ccb.dev.cube.errors import SwitchOperationError
from hvl_ccb.dev.cube.support import _SupportPort
from hvl_ccb.dev.cube.switches import _Switch
from masked_comm.opc import DemoServer


@pytest.fixture(scope="module")
def dev_config():
    return {
        "namespace_index": 3,
        "polling_delay_sec": 0.05,
        "polling_interval_sec": 0.01,
        "timeout_interval": 0.1,
        "timeout_status_change": 6,
        "noise_level_measurement_channel_1": 100,
        "noise_level_measurement_channel_2": 100,
        "noise_level_measurement_channel_3": 100,
        "noise_level_measurement_channel_4": 100,
    }


@pytest.fixture(scope="module")
def com_config():
    return {
        "host": "localhost",
        "port": 14123,
        "wait_timeout_retry_sec": 0.01,
        "max_timeout_retry_nr": 3,
    }


@pytest.fixture(scope="module")
def demo_opcua_server(dev_config):
    opcua_server = DemoServer(
        dev_config["namespace_index"], constants._CubeOpcEndpoint.BASE_CUBE, 14123
    )
    opcua_server.start()

    # add socket for CEE16
    opcua_server.set_var(constants._CEE16, False)
    # add T13 sockets
    for socket in constants._T13Socket._SOCKETS:
        opcua_server.set_var(constants._T13Socket(socket)._CMD, False)
    # add support input and output nodes
    for io, port, contact in product(
        _SupportPort._IOS,
        _SupportPort._PORTS,
        _SupportPort._CONTACTS,
    ):
        opcua_server.set_var(_SupportPort("", port)._cmd(io, contact), False)

    for i, alarm in enumerate(alarms._Alarms):
        opcua_server.set_var(alarm, bool(i % 2))

    # add remote control vars
    opcua_server.set_var(constants._OpcControl.ACTIVE, False)
    opcua_server.set_var(constants._OpcControl.LIVE, False)

    yield opcua_server

    opcua_server.stop()


@pytest.fixture(scope="module")
def opened_basecube_com(
    com_config,
    dev_config,  # noqa: ARG001
    demo_opcua_server: DemoServer,  # noqa: ARG001
):
    opc_comm = BaseCubeOpcUaCommunication(com_config)
    opc_comm.open()
    yield opc_comm
    opc_comm.close()


def _get_logged_messages(
    caplog,
    expected_logger_name: str,
    expected_level: int,
) -> list[str]:
    return [
        message
        for logger_name, level, message in caplog.record_tuples
        if logger_name == expected_logger_name and level == expected_level
    ]


def test_basecube_subscription_handler_datachange_safetystatus(
    opened_basecube_com,
    dev_config: dict,
    demo_opcua_server: DemoServer,
    caplog,  # pytest._logging.LogCaptureFixture
) -> None:
    ns_index = dev_config["namespace_index"]

    caplog.set_level(logging.INFO)

    # test a status datachange
    demo_opcua_server.set_var(
        constants._Safety.STATUS, int(constants.SafetyStatus.INITIALIZING)
    )
    opened_basecube_com.init_monitored_nodes(str(constants._Safety.STATUS), ns_index)
    sleep(0.05)

    demo_opcua_server.set_var(
        constants._Safety.STATUS,
        int(constants.SafetyStatus.GREEN_NOT_READY),
    )
    sleep(0.05)


def test_basecube_subscription_handler_datachange_earthingstickstatus(
    opened_basecube_com,
    dev_config: dict,
    demo_opcua_server: DemoServer,
    caplog,  # pytest._logging.LogCaptureFixture
) -> None:
    ns_index = dev_config["namespace_index"]

    caplog.set_level(logging.INFO)

    # test an earthing stick status datachange
    for i in earthing_stick._EarthingStick._STICKS:
        demo_opcua_server.set_var(
            earthing_stick._EarthingStick("", i)._CMD_STATUS,
            int(earthing_stick.SwitchStatus.INACTIVE),
        )
        opened_basecube_com.init_monitored_nodes(
            earthing_stick._EarthingStick("", i)._CMD_STATUS, ns_index
        )
    sleep(0.05)

    for i in earthing_stick._EarthingStick._STICKS:
        demo_opcua_server.set_var(
            earthing_stick._EarthingStick("", i)._CMD_STATUS,
            int(earthing_stick.SwitchStatus.CLOSED),
        )
    sleep(0.05)


@pytest.fixture(scope="module")
def cube(com_config, dev_config, demo_opcua_server: DemoServer):
    cube = BaseCube(com_config, dev_config)
    demo_opcua_server.set_var(constants._OpcControl.TIME, [0, 0, 0, 0, 0, 0, 0])
    cube.start()
    yield cube
    demo_opcua_server.set_var(
        constants._Safety.STATUS,
        int(constants.STOP_SAFETY_STATUSES[-1]),
    )
    cube.stop()


def test_dev_config(dev_config) -> None:
    # currently there are no non-default config values
    config = BaseCubeConfiguration(**dev_config)
    for key, value in dev_config.items():
        assert getattr(config, key) == value


@pytest.mark.parametrize(
    "wrong_config_dict",
    [
        {"namespace_index": -1},
        {"polling_interval_sec": -1},
        {"polling_interval_sec": 0},
        {"polling_delay_sec": -1},
        {"timeout_interval": 0},
        {"timeout_status_change": -1},
    ],
)
def test_invalid_config_dict_value_error(dev_config, wrong_config_dict) -> None:
    invalid_config = dict(dev_config)
    invalid_config.update(wrong_config_dict)
    with pytest.raises(ValueError):
        BaseCubeConfiguration(**invalid_config)


@pytest.mark.parametrize(
    "wrong_config_dict",
    [
        {"noise_level_measurement_channel_1": "100"},
    ],
)
def test_invalid_config_dict_type_error(dev_config, wrong_config_dict) -> None:
    invalid_config = dict(dev_config)
    invalid_config.update(wrong_config_dict)
    with pytest.raises(TypeError):
        BaseCubeConfiguration(**invalid_config)


def test_set_remote_control(cube: BaseCube, demo_opcua_server: DemoServer) -> None:
    cube._set_remote_control(False)
    assert demo_opcua_server.get_var(constants._OpcControl.ACTIVE) is False
    cube._set_remote_control(True)
    assert demo_opcua_server.get_var(constants._OpcControl.ACTIVE) is True

    with pytest.raises(TypeError):
        cube._set_remote_control(1)
    with pytest.raises(TypeError):
        cube._set_remote_control(0)
    with pytest.raises(TypeError):
        cube._set_remote_control("off")
    with pytest.raises(TypeError):
        cube._set_remote_control("on")

    demo_opcua_server.set_var(
        constants._Safety.STATUS, int(constants.SafetyStatus.RED_READY)
    )
    with pytest.raises(CubeRemoteControlError):
        cube._set_remote_control(False)
    demo_opcua_server.set_var(
        constants._Safety.STATUS, int(constants.SafetyStatus.GREEN_READY)
    )


def test_status_poller(
    com_config,
    dev_config,
    demo_opcua_server: DemoServer,
    mocker: MockerFixture,
) -> None:
    cube = BaseCube(com_config, dev_config)
    poller = cube._status_poller
    assert poller is not None
    assert poller.polling_delay_sec == dev_config["polling_delay_sec"]
    assert poller.polling_interval_sec == dev_config["polling_interval_sec"]
    assert poller.polling_timeout_sec is None
    assert poller.spoll_handler == cube._spoll_handler

    spy_spoll_handler = mocker.spy(poller, "spoll_handler")
    spy_start_polling = mocker.spy(poller, "start_polling")
    spy_stop_polling = mocker.spy(poller, "stop_polling")

    poller.stop_polling()
    assert spy_stop_polling.spy_return is False
    demo_opcua_server.set_var(constants._OpcControl.TIME, [0, 0, 0, 0, 0, 0, 0])
    cube.start()

    spy_start_polling.assert_called()
    assert spy_start_polling.spy_return is True

    poller.start_polling()
    assert spy_start_polling.spy_return is False

    sleep(dev_config["polling_delay_sec"] + 3 * dev_config["polling_interval_sec"])
    spy_spoll_handler.assert_called()
    assert spy_spoll_handler.spy_return is None

    cube.stop()
    spy_stop_polling.assert_called()
    assert spy_stop_polling.spy_return is True


def test_status_poller_timeout_error(
    com_config,
    dev_config,
    demo_opcua_server: DemoServer,
    mocker: MockerFixture,
) -> None:
    cube = BaseCube(com_config, dev_config)
    demo_opcua_server.set_var(constants._OpcControl.TIME, [0, 0, 0, 0, 0, 0, 0])
    cube.start()
    assert cube._status_poller.is_polling()

    # patch UASocketProtocol.send_request to raise a mock TimeoutError as if coming from
    # used therein concurrent.futures.Future;
    # raise error only on disable remote control done via poller thread on stop()
    # => write(constants.OpcControl.active, False)

    # Use bound method (otherwise live unpatch does not work):
    send_request_orig = cube.com._client.uaclient.protocol.send_request

    async def send_request(_self, request, *args, **kwargs):
        if hasattr(request, "Parameters") and hasattr(
            request.Parameters, "NodesToWrite"
        ):
            node_to_write = request.Parameters.NodesToWrite[0]
            if (
                node_to_write.NodeId.Identifier == constants._OpcControl.ACTIVE
                and node_to_write.Value.Value.Value is False
            ):
                if sys.version_info < (3, 11):
                    from concurrent.futures import TimeoutError as FuturesTimeoutError
                else:
                    FuturesTimeoutError = TimeoutError  # noqa: N806

                msg = "mock timeout error"
                raise FuturesTimeoutError(msg)
        # method already bound - ignore `self`
        return await send_request_orig(request, *args, **kwargs)

    mocker.patch(
        "asyncua.client.ua_client.UASocketProtocol.send_request",
        side_effect=send_request,
        autospec=True,
    )

    # check poller thread error caught and wrapped
    with pytest.raises(OpcUaCommunicationTimeoutError):
        cube._set_remote_control(False)
    assert not cube._status_poller.is_polling()

    # check that stopping also does raises the error, but cleans up otherwise
    cube._set_remote_control(True)
    with pytest.raises(OpcUaCommunicationTimeoutError):
        cube.stop()
    assert not cube._status_poller.is_polling()

    # unpatch and try to stop dev again - will raise error due to broken com,
    # but not the timeout
    mocker.patch(
        "asyncua.client.ua_client.UASocketProtocol.send_request",
        side_effect=send_request_orig,
    )

    with pytest.raises(OpcUaCommunicationIOError) as excinfo:
        cube.stop()
    assert excinfo.type is not OpcUaCommunicationTimeoutError
    assert not cube._status_poller.is_polling()


def test_support_input(cube: BaseCube) -> None:
    assert cube.support_1.input_1 is False
    assert cube.support_1.input_2 is False


def test_support_output(cube: BaseCube) -> None:
    cube.support_1.output_1 = False
    assert cube.support_1.output_1 is False
    cube.support_1.output_1 = True
    assert cube.support_1.output_1 is True

    cube.support_1.output_2 = False
    assert cube.support_1.output_2 is False
    cube.support_1.output_2 = True
    assert cube.support_1.output_2 is True

    with pytest.raises(TypeError):
        cube.support_1.output_1 = 1
    with pytest.raises(TypeError):
        cube.support_1.output_1 = 0
    with pytest.raises(TypeError):
        cube.support_1.output_1 = "off"
    with pytest.raises(TypeError):
        cube.support_1.output_1 = "on"


def test_t13_socket(cube: BaseCube) -> None:
    assert cube.t13_socket_1 is False
    cube.t13_socket_1 = True
    assert cube.t13_socket_1 is True
    with pytest.raises(TypeError):
        cube.t13_socket_1 = 1
    with pytest.raises(TypeError):
        cube.t13_socket_1 = 0
    with pytest.raises(TypeError):
        cube.t13_socket_1 = "off"
    with pytest.raises(TypeError):
        cube.t13_socket_1 = "on"


def test_cee16_socket(cube: BaseCube) -> None:
    cube.cee16_socket = False
    assert cube.cee16_socket is False
    cube.cee16_socket = True
    assert cube.cee16_socket is True

    with pytest.raises(TypeError):
        cube.cee16_socket = 1
    with pytest.raises(TypeError):
        cube.cee16_socket = 0
    with pytest.raises(TypeError):
        cube.cee16_socket = "off"
    with pytest.raises(TypeError):
        cube.cee16_socket = "on"


def test_get_status(cube: BaseCube, demo_opcua_server: DemoServer) -> None:
    demo_opcua_server.set_var(constants._Safety.STATUS, 1)
    assert cube.status == constants.SafetyStatus.GREEN_NOT_READY

    demo_opcua_server.set_var(constants._Safety.STATUS, 2)
    assert cube.status == constants.SafetyStatus.GREEN_READY


def _switch_safety_status(
    demo_opcua_server: DemoServer,
    safety_status: constants.SafetyStatus,
    delay: float = BaseCubeConfiguration.timeout_interval * 2,
):
    sleep(delay)
    demo_opcua_server.set_var(constants._Safety.STATUS, int(safety_status))


def test_ready(cube: BaseCube, demo_opcua_server: DemoServer) -> None:
    demo_opcua_server.set_var(
        constants._SafetyStatusTransition.SWITCH_TO_READY.command, False
    )
    demo_opcua_server.set_var(
        constants._Safety.STATUS, int(constants.SafetyStatus.GREEN_READY)
    )
    executor = ThreadPoolExecutor(max_workers=1)
    executor.submit(
        _switch_safety_status,
        demo_opcua_server,
        constants.SafetyStatus.RED_READY,
    )
    cube.ready = True
    assert (
        demo_opcua_server.get_var(
            constants._SafetyStatusTransition.SWITCH_TO_READY.command
        )
        is True
    )
    executor.submit(
        _switch_safety_status,
        demo_opcua_server,
        constants.SafetyStatus.GREEN_READY,
    )
    assert cube.ready is True
    cube.ready = False
    assert (
        demo_opcua_server.get_var(
            constants._SafetyStatusTransition.SWITCH_TO_READY.command
        )
        is False
    )
    assert cube.ready is False
    demo_opcua_server.set_var(
        constants._Safety.STATUS, int(constants.SafetyStatus.RED_OPERATE)
    )
    assert cube.ready is None


def test_operate(cube: BaseCube, demo_opcua_server: DemoServer) -> None:
    demo_opcua_server.set_var(
        constants._SafetyStatusTransition.SWITCH_TO_OPERATE.command, False
    )
    demo_opcua_server.set_var(
        constants._Safety.STATUS, int(constants.SafetyStatus.RED_READY)
    )
    executor = ThreadPoolExecutor(max_workers=1)
    executor.submit(
        _switch_safety_status,
        demo_opcua_server,
        constants.SafetyStatus.RED_OPERATE,
    )
    cube.operate = True
    assert (
        demo_opcua_server.get_var(
            constants._SafetyStatusTransition.SWITCH_TO_OPERATE.command
        )
        is True
    )
    executor.submit(
        _switch_safety_status,
        demo_opcua_server,
        constants.SafetyStatus.RED_READY,
    )
    assert cube.operate is True
    cube.operate = False
    assert (
        demo_opcua_server.get_var(
            constants._SafetyStatusTransition.SWITCH_TO_OPERATE.command
        )
        is False
    )
    assert cube.operate is False
    demo_opcua_server.set_var(
        constants._Safety.STATUS, int(constants.SafetyStatus.GREEN_READY)
    )
    assert cube.operate is None


def test_switch_safety_status(cube: BaseCube, demo_opcua_server: DemoServer) -> None:
    demo_opcua_server.set_var(
        constants._Safety.STATUS, int(constants.SafetyStatus.GREEN_NOT_READY)
    )
    with pytest.raises(CubeStatusChangeError):
        cube._switch_safety_status(
            constants._SafetyStatusTransition.SWITCH_TO_READY, True
        )
    with pytest.raises(CubeStatusChangeError):
        cube._switch_safety_status(
            constants._SafetyStatusTransition.SWITCH_TO_READY, False
        )
    demo_opcua_server.set_var(
        constants._Safety.STATUS, int(constants.SafetyStatus.GREEN_READY)
    )
    demo_opcua_server.set_var(
        constants._SafetyStatusTransition.SWITCH_TO_READY.command, False
    )
    with pytest.raises(CubeStatusChangeError):
        cube._switch_safety_status(
            constants._SafetyStatusTransition.SWITCH_TO_READY, True
        )


def test_get_measurement_ratio(cube: BaseCube, demo_opcua_server: DemoServer) -> None:
    demo_opcua_server.set_var(
        constants._MeasurementChannel("", 1, 400)._CMD_RATIO, 123.4
    )
    assert cube.measurement_ch_1.ratio == 123.4


def test_get_measurement_voltage(cube: BaseCube, demo_opcua_server: DemoServer) -> None:
    demo_opcua_server.set_var(
        constants._MeasurementChannel("", 1, 400)._CMD_SCALE, True
    )
    demo_opcua_server.set_var(
        constants._MeasurementChannel("", 1, 400)._CMD_VOLTAGE, 110
    )
    assert cube.measurement_ch_1.voltage == 110_000.0
    demo_opcua_server.set_var(
        constants._MeasurementChannel("", 1, 400)._CMD_SCALE, False
    )
    demo_opcua_server.set_var(
        constants._MeasurementChannel("", 1, 400)._CMD_VOLTAGE, 110
    )
    assert cube.measurement_ch_1.voltage == 110


def test_switch_without_plc_fields(
    cube: BaseCube,
    demo_opcua_server: DemoServer,  # noqa: ARG001
) -> None:
    class Switch(_Switch):
        """Inheritance of _Switch as this cannot be instanced"""

    switch = Switch(cube, "wrong_switch")

    with pytest.raises(SwitchOperationError):
        assert switch.status

    with pytest.raises(SwitchOperationError):
        assert switch.operating_status

    with pytest.raises(SwitchOperationError):
        assert switch.operate


def test_get_earthing_stick_status(
    cube: BaseCube, demo_opcua_server: DemoServer
) -> None:
    demo_opcua_server.set_var(earthing_stick._EarthingStick("", 1)._CMD_STATUS, 1)
    assert cube.earthing_stick_1.status == earthing_stick.SwitchStatus(1)


def test_get_earthing_stick_operating_status(
    cube: BaseCube, demo_opcua_server: DemoServer
) -> None:
    demo_opcua_server.set_var(
        earthing_stick._EarthingStick("", 1)._CMD_OPERATING_STATUS,
        earthing_stick.SwitchOperatingStatus.AUTO,
    )
    assert (
        cube.earthing_stick_1.operating_status
        is earthing_stick.SwitchOperatingStatus.AUTO
    )


def test_get_earthing_stick_manual(
    cube: BaseCube, demo_opcua_server: DemoServer
) -> None:
    demo_opcua_server.set_var(
        earthing_stick._EarthingStick("", 1)._CMD_MANUAL,
        bool(earthing_stick.SwitchOperation.CLOSE),
    )
    assert cube.earthing_stick_1.operate is earthing_stick.SwitchOperation.CLOSE


def test_operate_earthing_stick(cube: BaseCube, demo_opcua_server: DemoServer) -> None:
    demo_opcua_server.set_var(
        earthing_stick._EarthingStick("", 1)._CMD_OPERATING_STATUS,
        earthing_stick.SwitchOperatingStatus.MANUAL,
    )
    demo_opcua_server.set_var(
        earthing_stick._EarthingStick("", 1)._CMD_MANUAL,
        bool(earthing_stick.SwitchOperation.OPEN),
    )

    # Tests @ READ_READY
    demo_opcua_server.set_var(
        constants._Safety.STATUS, int(constants.SafetyStatus.RED_READY)
    )
    assert cube.earthing_stick_1.operate is earthing_stick.SwitchOperation.OPEN
    cube.earthing_stick_1.operate = earthing_stick.SwitchOperation.CLOSE
    assert cube.earthing_stick_1.operate is earthing_stick.SwitchOperation.CLOSE
    cube.earthing_stick_1.operate = False
    assert cube.earthing_stick_1.operate is earthing_stick.SwitchOperation.OPEN

    # Tests @ READ_OPERATE
    demo_opcua_server.set_var(
        constants._Safety.STATUS, int(constants.SafetyStatus.RED_OPERATE)
    )
    cube.earthing_stick_1.operate = earthing_stick.SwitchOperation.OPEN
    assert cube.earthing_stick_1.operate is earthing_stick.SwitchOperation.OPEN
    cube.earthing_stick_1.operate = earthing_stick.SwitchOperation.CLOSE
    assert cube.earthing_stick_1.operate is earthing_stick.SwitchOperation.CLOSE

    # No operation @ GREEN_READY
    demo_opcua_server.set_var(
        constants._Safety.STATUS, int(constants.SafetyStatus.GREEN_READY)
    )
    with pytest.raises(CubeEarthingStickOperationError):
        cube.earthing_stick_1.operate = True

    # No operation in non-manual mode
    demo_opcua_server.set_var(
        constants._Safety.STATUS, int(constants.SafetyStatus.RED_READY)
    )
    demo_opcua_server.set_var(
        earthing_stick._EarthingStick("", 1)._CMD_OPERATING_STATUS,
        earthing_stick.SwitchOperatingStatus.AUTO,
    )
    with pytest.raises(CubeEarthingStickOperationError):
        cube.earthing_stick_1.operate = True


def test_operate_earthing_stick_error(
    cube: BaseCube,
    demo_opcua_server: DemoServer,
) -> None:
    demo_opcua_server.set_var(
        earthing_stick._EarthingStick("", 1)._CMD_OPERATING_STATUS,
        earthing_stick.SwitchOperatingStatus.AUTO,
    )
    with pytest.raises(CubeEarthingStickOperationError):
        cube.earthing_stick_1.operate = earthing_stick.SwitchOperation.CLOSE


def test_quit_error(cube: BaseCube, demo_opcua_server: DemoServer) -> None:
    demo_opcua_server.set_var(constants._Errors.QUIT, False)
    cube.quit_error()
    assert demo_opcua_server.get_var(constants._Errors.QUIT) is False


def test_get_door_status(cube: BaseCube, demo_opcua_server: DemoServer) -> None:
    demo_opcua_server.set_var(constants._Door(1, "")._CMD, 2)
    assert cube.door_1_status == constants.DoorStatus(2)


def test_base_descriptor(cube: BaseCube, demo_opcua_server: DemoServer) -> None:  # noqa: ARG001
    with pytest.raises(AttributeError):
        cube.door_1_status = True


def test_get_earthing_rod_status(cube: BaseCube, demo_opcua_server: DemoServer) -> None:
    demo_opcua_server.set_var(constants._EarthingRod(1, "")._CMD, 0)
    assert cube.earthing_rod_1_status == constants.EarthingRodStatus(0)
    demo_opcua_server.set_var(constants._EarthingRod(1, "")._CMD, 1)
    assert cube.earthing_rod_1_status == constants.EarthingRodStatus(1)


def test_get_breakdown_detection_active(
    cube: BaseCube, demo_opcua_server: DemoServer
) -> None:
    demo_opcua_server.set_var(constants._BreakdownDetection.ACTIVATED, True)
    assert cube.breakdown_detection_active is True


def test_get_breakdown_detection_triggered(
    cube: BaseCube, demo_opcua_server: DemoServer
) -> None:
    demo_opcua_server.set_var(constants._BreakdownDetection.TRIGGERED, True)
    assert cube.breakdown_detection_triggered is True


def test_breakdown_detection_reset(
    cube: BaseCube, demo_opcua_server: DemoServer
) -> None:
    demo_opcua_server.set_var(constants._BreakdownDetection.RESET, False)
    cube.breakdown_detection_reset()
    assert demo_opcua_server.get_var(constants._BreakdownDetection.RESET) is False


def test_set_status_board(cube: BaseCube, demo_opcua_server: DemoServer) -> None:
    n_max = len(constants.MessageBoard)

    with pytest.raises(ValueError):
        cube.set_status_board(["x"] * (n_max + 1))
    with pytest.raises(ValueError):
        cube.set_status_board(["x"], pos=[n_max])

    for line in constants.MessageBoard:
        demo_opcua_server.set_var(line, "")

    msgs = ["Hello World", "Hello Fabian", "Hello HVL"]

    cube.set_status_board(msgs[:1])
    assert demo_opcua_server.get_var(constants.MessageBoard.line(1)) == msgs[0]
    cube.set_status_board(msgs[:1], pos=[1], clear_board=False)
    assert demo_opcua_server.get_var(constants.MessageBoard.line(1)) == msgs[0]
    assert demo_opcua_server.get_var(constants.MessageBoard.line(2)) == msgs[0]

    positions = [4, 8, 12]
    cube.set_status_board(msgs, positions, clear_board=False)
    for i, p in enumerate(positions):
        assert demo_opcua_server.get_var(constants.MessageBoard.line(p + 1)) == msgs[i]

    cube.set_status_board([str(i) for i in range(15)])
    for i in range(15):
        assert demo_opcua_server.get_var(constants.MessageBoard.line(i + 1)) == str(i)


def test_set_message_board(cube: BaseCube, demo_opcua_server: DemoServer) -> None:
    n_max = len(constants.MessageBoard)

    with pytest.raises(ValueError):
        cube.set_message_board(["x"] * (n_max + 1))

    for line in constants.MessageBoard:
        demo_opcua_server.set_var(line, "")

    msgs = ["ERROR: This is unexpected", "Good choice"]
    cube.set_message_board(msgs[:1])
    assert demo_opcua_server.get_var(constants.MessageBoard.line(1)).endswith(msgs[0])

    # push two messages to first two lines, in the given order
    cube.set_message_board(msgs)
    assert demo_opcua_server.get_var(constants.MessageBoard.line(1)).endswith(msgs[0])
    assert demo_opcua_server.get_var(constants.MessageBoard.line(2)).endswith(msgs[1])
    assert demo_opcua_server.get_var(constants.MessageBoard.line(3)).endswith(msgs[0])

    # overwrite all messages except last, which is pushed to be last line
    cube.set_message_board("x" * (n_max - 1))
    for n in range(len(constants.MessageBoard) - 1):
        assert demo_opcua_server.get_var(constants.MessageBoard.line(n + 1)).endswith(
            "x"
        )
    last_msg = demo_opcua_server.get_var(constants.MessageBoard.line(n_max))
    assert last_msg.endswith(msgs[0])


def test_active_alarms(cube: BaseCube, demo_opcua_server: DemoServer) -> None:  # noqa: ARG001
    cube._com.config.sub_handler.alarm_status = alarms._AlarmsOverview()
    cube._com.config.sub_handler.alarm_status.alarm_50 = alarms._AlarmStatus.ACTIVE
    assert cube.active_alarms(False) == [50]
    assert cube.active_alarms(True) == ["OPC Connection active"]


def test_stop(cube: BaseCube, demo_opcua_server: DemoServer) -> None:
    demo_opcua_server.set_var(
        constants._Safety.STATUS, int(constants.SafetyStatus.RED_READY)
    )
    with pytest.raises(CubeStopError):
        cube.stop()
